#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import mimiqcircuits as mc
from mimiqcircuits.operations.operation import Operation
from mimiqcircuits.operations.measure import AbstractMeasurement


class MeasureReset(AbstractMeasurement):
    """MeasureReset operation.

    This operation measures a qubit q, stores the value in a classical bit c,
    then applies a X operation to the qubit if the measured value is 1, effectively
    resetting the qubit to the :math:`\\ket{0}` state.

    Examples:

        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(MeasureReset(), 1, 0)
        2-qubit circuit with 1 instructions:
        └── MR @ q[1], c[0]
        <BLANKLINE>
        >>> c.decompose()
        2-qubit circuit with 2 instructions:
        ├── M @ q[1], c[0]
        └── IF (c==1) X @ q[1], c[0]
        <BLANKLINE>
    """

    _name = "MR"
    _num_bits = 1
    _num_qubits = 1
    _num_cregs = 1
    _qregsizes = [1]
    _cregsizes = [1]
    _num_zvars = 0

    def inverse(self):
        raise TypeError("MeasureReset is not inversible")

    def power(self, p):
        raise TypeError("MeasureReset^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled MeasureReset is not defined.")

    def iswrapper(self):
        return False

    def _decompose(self, circ, qubits, bits, zvars):
        circ.push(mc.Measure(), qubits, bits)
        circ.push(mc.IfStatement(mc.GateX(), bitstring=mc.BitString("1")), qubits, bits)

        return circ

    def get_operation(self):
        return self

    def __str__(self):
        return f"{self._name}"


class MeasureResetX(AbstractMeasurement):
    """MeasureResetX operation.

    The MeasureResetX operation first applies a Hadamard gate (H) to the qubit,
    performs a measurement and reset operation similar to the MeasureReset operation,
    and then applies another Hadamard gate. This sequence effectively measures the
    qubit in the X-basis and resets it to the :math:`|+>` state.

    See Also:
        :class:`MeasureReset`: The standard measurement and reset operation in the Z-basis.
        :class:`MeasureResetY`: Similar operation that measures and resets the qubit in the Y-basis.

    Examples:

        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(MeasureResetX(), 1, 0)
        2-qubit circuit with 1 instructions:
        └── MRX @ q[1], c[0]
        <BLANKLINE>
        >>> c.decompose()
        2-qubit circuit with 3 instructions:
        ├── H @ q[1]
        ├── MR @ q[1], c[0]
        └── H @ q[1]
        <BLANKLINE>
    """

    _name = "MRX"
    _num_bits = 1
    _num_qubits = 1
    _num_cregs = 1
    _qregsizes = [1]
    _cregsizes = [1]
    _num_zvars = 0

    def inverse(self):
        raise TypeError("MeasureResetX is not inversible")

    def power(self, p):
        raise TypeError("MeasureResetX^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled MeasureResetX is not defined.")

    def iswrapper(self):
        return False

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        c = bits[0]
        circ.push(mc.GateH(), q)
        circ.push(MeasureReset(), q, c)
        circ.push(mc.GateH(), q)
        return circ

    def get_operation(self):
        return self

    def __str__(self):
        return f"{self._name}"


class MeasureResetY(AbstractMeasurement):
    """MeasureResetY operation.

    The MeasureResetY operation applies (HYZ) gate to
    the qubit, performs a MeasureReset operation,
    and then applies another HYZ gate. 
    This sequence effectively measures the qubit in
    the Y-basis.

    See Also:
        :class:`MeasureReset`: The standard measurement and reset operation in the Z-basis.
        :class:`MeasureResetX`: Similar operation that measures and resets the qubit in the X-basis.

    Examples:

        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(MeasureResetY(), 1, 0)
        2-qubit circuit with 1 instructions:
        └── MRY @ q[1], c[0]
        <BLANKLINE>
        >>> c.decompose()
        2-qubit circuit with 3 instructions:
        ├── HYZ @ q[1]
        ├── MR @ q[1], c[0]
        └── HYZ @ q[1]
        <BLANKLINE>
    """

    _name = "MRY"
    _num_bits = 1
    _num_qubits = 1
    _num_cregs = 1
    _qregsizes = [1]
    _cregsizes = [1]
    _num_zvars = 0

    def inverse(self):
        raise TypeError("MeasureResetY is not inversible")

    def power(self, p):
        raise TypeError("MeasureResetY^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled MeasureResetY is not defined.")

    def iswrapper(self):
        return False

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        c = bits[0]
        circ.push(mc.GateHYZ(), q)
        circ.push(MeasureReset(), q, c)
        circ.push(mc.GateHYZ(), q)
        return circ

    def get_operation(self):
        return self

    def __str__(self):
        return f"{self._name}"

class MeasureResetZ(AbstractMeasurement):
    r"""This class  acting as an alias for :class:`MeasureReset`.
    """
    def __new__(self):
        return MeasureReset()


# export operations
__all__ = ["MeasureReset", "MeasureResetX", "MeasureResetY",  "MeasureResetZ"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.gates.gate as mcg
from mimiqcircuits.operations.gates.standard.u import GateU
from mimiqcircuits.matrices import pmatrix, umatrix
from symengine import pi


class GateU1(mcg.Gate):
    r"""Single qubit generic unitary gate :math:`{U_1}`.

    Equivalent to :func:`GateP`

    **Matrix representation:**

    .. math::
        \operatorname{U1}(\lambda) = \begin{pmatrix}
            1 & 0 \\
            0 & e^{i\lambda}
        \end{pmatrix}

    Parameters:
        lambda (float): Euler angle 3 in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> lmbda = Symbol('lambda')
        >>> GateU1(lmbda)
        U1(lambda)
        >>> GateU1(lmbda).matrix()
        [1.0, 0]
        [0, exp(I*lambda)]
        <BLANKLINE>
        >>> c = Circuit().push(GateU1(lmbda), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── U1(lambda) @ q[0]
        <BLANKLINE>
        >>> GateU1(lmbda).power(2), GateU1(lmbda).inverse()
        (U1(2*lambda), U1(-lambda))
        >>> GateU1(lmbda).decompose()
        1-qubit circuit with 1 instructions:
        └── U(0, 0, lambda, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "U1"

    _num_qubits = 1
    _qregsizes = [1]

    _parnames = ("lmbda",)

    def __init__(self, lmbda):
        self.lmbda = lmbda

    def _matrix(self):
        return pmatrix(self.lmbda)

    def inverse(self):
        return GateU1(-self.lmbda)

    def _power(self, pwr):
        return GateU1(pwr * self.lmbda)

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(0, 0, self.lmbda), q)
        return circ


class GateU2(mcg.Gate):
    r"""Single qubit generic unitary gate :math:`{U_2}`.


    **Matrix representation:**

    .. math::
        \operatorname{U2}(\phi,\lambda) = \frac{1}{\sqrt{2}}e^{-(\phi+\lambda)/2}\begin{pmatrix}
            1 & -e^{i\lambda} \\
            e^{i\phi} & e^{i(\phi+\lambda)}
        \end{pmatrix}

    Parameters:
        phi: Euler angle in radians.
        lambda: Euler angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> phi, lmbda = symbols('phi lambda')
        >>> GateU2(phi, lmbda)
        U2(phi, lambda)
        >>> GateU2(phi, lmbda).matrix()
        [0.707106781186548, -0.707106781186548*exp(I*lambda)]
        [0.707106781186548*exp(I*phi), 0.707106781186548*exp(I*(lambda + phi))]
        <BLANKLINE>
        >>> c = Circuit().push(GateU2(phi, lmbda), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── U2(phi, lambda) @ q[0]
        <BLANKLINE>
        >>> GateU2(phi, lmbda).power(2), GateU2(phi, lmbda).inverse()
        (U2(phi, lambda)**2, U2(-lambda - pi, -phi + pi))
        >>> GateU2(phi, lmbda).decompose()
        1-qubit circuit with 1 instructions:
        └── U((1/2)*pi, phi, lambda, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "U2"

    _num_qubits = 1
    _qregsizes = [1]

    _parnames = ("phi", "lmbda")

    def __init__(self, phi, lmbda):
        self.phi = phi
        self.lmbda = lmbda

    def _matrix(self):
        return umatrix(pi / 2, self.phi, self.lmbda)

    def inverse(self):
        return GateU2(-self.lmbda - pi, -self.phi + pi)

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(pi / 2, self.phi, self.lmbda), q)
        return circ


class GateU3(mcg.Gate):
    r"""Single qubit generic unitary gate :math:`{U_3}`.

    This gate is equivalent to :func:`GateU` up to a global phase, :math:`\operatorname{U3}(\theta,\phi,\lambda) = e^{-i(\phi + \lambda + \theta)/2} \operatorname{U}(\theta,\phi,\lambda)` 

    **Matrix representation:**

    .. math::
        \operatorname{U3}(\theta,\phi,\lambda) = \frac{1}{2}e^{-i(\phi + \lambda + \theta)/2}
        \begin{pmatrix}
        1 + e^{i\theta} & -i e^{i\lambda}(1 - e^{i\theta}) \\
        i e^{i\phi}(1 - e^{i\theta}) & e^{i(\phi + \lambda)}(1 + e^{i\theta})
        \end{pmatrix}

    Parameters:
        theta: Euler angle 1 in radians.
        phi: Euler angle 2 in radians.
        lambda: Euler angle 3 in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta, phi, lmbda = symbols('theta phi lambda')
        >>> GateU3(theta, phi, lmbda)
        U3(theta, phi, lambda)
        >>> GateU3(theta, phi, lmbda).matrix()
        [1.0*cos((1/2)*theta), -exp(I*lambda)*sin((1/2)*theta)]
        [exp(I*phi)*sin((1/2)*theta), exp(I*(lambda + phi))*cos((1/2)*theta)]
        <BLANKLINE>
        >>> c = Circuit().push(GateU3(theta, phi, lmbda), 0)
        >>> GateU3(theta, phi, lmbda).power(2), GateU3(theta, phi, lmbda).inverse()
        (U3(theta, phi, lambda)**2, U3(-theta, -lambda, -phi))
        >>> GateU3(theta, phi, lmbda).decompose()
        1-qubit circuit with 1 instructions:
        └── U(theta, phi, lambda, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "U3"

    _num_qubits = 1
    _qregsizes = [1]

    _parnames = ("theta", "phi", "lmbda")

    def __init__(self, theta, phi, lmbda):
        self.theta = theta
        self.phi = phi
        self.lmbda = lmbda

    def _matrix(self):
        return umatrix(self.theta, self.phi, self.lmbda)

    def inverse(self):
        return GateU3(-self.theta, -self.lmbda, -self.phi)

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(self.theta, self.phi, self.lmbda), q)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.operation import Operation
import mimiqcircuits as mc
import copy
import numpy as np


def _allunique(lst):
    seen = set()
    for element in lst:
        if element in seen:
            return False
        seen.add(element)
    return True


class Instruction:
    """Initializes an instruction of a quantum circuit.

    Args:
        operation (Operation): The operation applied by the instruction.
        qubits (tuple of int): The qubits to apply the quantum operation to.
        bits (tuple of int): The classical bits to apply the quantum operation to.

    Raises:
        TypeError: If operation is not a subclass of Gate or qubits is not a tuple.
        ValueError: If qubits contains less than 1 or more than 2 elements.

    Examples:

        >>> from mimiqcircuits import *
        >>> Instruction(GateX(),(0,),())
        X @ q[0]
        >>> Instruction(Barrier(4),(0,1,2,3),())
        Barrier @ q[0,1,2,3]

    """

    _operation = None
    _qubits = None
    _bits = None
    _zvars = None

    def __init__(self, operation, qubits=None, bits=None, zvars=None):
        if qubits is None:
            qubits = tuple()

        if bits is None:
            bits = tuple()

        if zvars is None:
            zvars = tuple()

        if not isinstance(qubits, tuple):
            raise TypeError(
                f"Target qubits should be given in a tuple of integers. Given {qubits} of type {type(qubits)}."
            )

        if not isinstance(bits, tuple):
            raise TypeError(
                f"Target bits should be given in a tuple of integers. Given {bits} of type {type(bits)}."
            )

        if not isinstance(zvars, tuple):
            raise TypeError(
                f"Target z-variables should be given in a tuple of integers. Given {zvars} of type {type(zvars)}."
            )

        if not isinstance(operation, Operation):
            raise TypeError(
                f"Operation must be a subclass of Operation. Given {operation} of type f{type(operation)}"
            )

        if not _allunique(qubits):
            raise ValueError("Duplicated qubit target in instruction")

        if not _allunique(bits):
            raise ValueError("Duplicated classical bit target in instruction")

        if not _allunique(zvars):
            raise ValueError("Duplicated z-variables target in instruction")

        for qi in qubits:
            if qi < 0:
                raise ValueError("Qubit target index cannot be negative")

        for bi in bits:
            if bi < 0:
                raise ValueError("Bit target index cannot be negative")

        for z in zvars:
            if z < 0:
                raise ValueError("Z-variable target index cannot be negative")

        if len(qubits) != operation.num_qubits:
            raise ValueError(
                f"Wrong number of target qubits for operation {operation} wanted  {operation.num_qubits}, given {len(qubits)}"
            )

        if len(bits) != operation.num_bits:
            raise ValueError(
                f"Wrong number of target bits for operation {operation} wanted  {operation.num_bits}, given {len(bits)}"
            )

        if len(zvars) != operation.num_zvars:
            raise ValueError(
                f"Wrong number of z-targets for operation {operation} wanted  {operation.num_zvars}, given {len(zvars)}"
            )

        self._operation = operation
        self._qubits = qubits
        self._bits = bits
        self._zvars = zvars

    @property
    def operation(self):
        return self._operation

    @operation.setter
    def operation(self, _):
        raise AttributeError("operation is a read-only attribute")

    @property
    def qubits(self):
        return self._qubits

    @qubits.setter
    def qubits(self, _):
        raise AttributeError("qubits is a read-only attribute")

    @property
    def bits(self):
        return self._bits

    @bits.setter
    def bits(self, _):
        raise AttributeError("bits is a read-only attribute")

    @property
    def zvars(self):
        return self._zvars

    @zvars.setter
    def zvars(self, _):
        raise AttributeError("zvars is a read-only attribute")

    def __repr__(self):
        return str(self)

    def get_operation(self):
        return self._operation

    def get_qubits(self):
        return self._qubits

    def get_bits(self):
        return self._bits

    def get_zvars(self):
        return self._zvars

    def num_qubits(self):
        return len(self._qubits)

    def num_bits(self):
        return len(self._bits)

    def num_zvars(self):
        return len(self._zvars)

    def get_operation(self):
        return self.operation

    def asciiwidth(self):
        return self.operation.asciiwidth(self._qubits, self._bits)

    def __eq__(self, other):
        if not isinstance(other, Instruction):
            return False
        return (
            (self.operation == other.operation)
            and (self.qubits == other.qubits)
            and (self.bits == other.bits)
            and (self.zvars == other.zvars)
        )

    def inverse(self):
        return Instruction(self.operation.inverse(), self.qubits, self.bits)

    def copy(self):
        """Creates a shallow copy of the instruction.
            To create a full copy use deepcopy() instead.

        Returns:
            Instruction: A new Instruction object containing references to the same attributes as the original circuit
        """
        return copy.copy(self)

    def deepcopy(self):
        """Creates a copy of the object and for all its attributes

        Returns:
            Instruction: A new Instruction object fully identical the original circuit
        """
        return copy.deepcopy(self)

    def _decompose(self, circ):
        return self.operation._decompose(circ, self.qubits, self.bits, self.zvars)

    def decompose(self):
        return self._decompose(mc.Circuit())

    def evaluate(self, d):
        return Instruction(
            self.operation.evaluate(d), self.qubits, self.bits, self.zvars
        )

    def __str__(self):
        compact = False
        op = str(self.operation)
        nq = len(self.qubits)
        nb = len(self.bits)
        nz = len(self.zvars)
        space = "" if compact else " "
        targets = ""
        if nq != 0 or nb != 0 or nz != 0:
            targets = f"{space}@{space}"

            if nq != 0:
                q_partition = _partition(
                    self.get_qubits(), np.cumsum(self.operation.qregsizes)
                )
                q_targets = f",{space}".join(
                    f"q{_string_with_square(_find_unit_range(x), ',')}"
                    for x in q_partition
                )
                targets += f",".join(q_targets.split(","))

            if nb != 0:
                if nq != 0:  # If qubits exist, add a separator before classical bits
                    targets += f", "
                c_partition = _partition(
                    self.get_bits(), np.cumsum(self.operation.cregsizes)
                )
                c_targets = "".join(  # Concatenate classical bits without extra commas
                    f"c{_string_with_square(x, ',')}" for x in c_partition
                )
                targets += c_targets

            if nz != 0:
                if nq != 0 or nb != 0:  # Add a separator before z-vars only if qubits or bits exist
                    targets += f", "
                z_partition = _partition(
                    self.get_zvars(), np.cumsum(self.operation.zregsizes)
                )
                z_targets = "".join(  # Concatenate z-vars without extra commas
                    f"z{_string_with_square(x, ',')}" for x in z_partition
                )
                targets += z_targets

        return f"{op}{targets}"


def _partition(arr, indices):
    vec = list(arr)
    partitions = [vec[: indices[0]]]

    for i in range(1, len(indices)):
        partitions.append(vec[indices[i - 1]: indices[i]])

    return partitions


def _string_with_square(arr, sep):
    return (
        "["
        + sep.join(
            map(lambda e: sep.join(map(str, e))
                if isinstance(e, list) else str(e), arr)
        )
        + "]"
    )


def _find_unit_range(arr):
    if len(arr) < 2:
        return arr

    narr = []
    rangestart = arr[0]
    rangestop = arr[0]

    for v in arr[1:]:
        if v == rangestop + 1:
            rangestop = v
        elif rangestart == rangestop:
            narr.append(rangestart)
            rangestart = v
            rangestop = v
        else:
            narr.append(list(range(rangestart, rangestop + 1)))
            rangestart = v
            rangestop = v

    if rangestart == rangestop:
        narr.append(rangestart)
    else:
        narr.append(list(range(rangestart, rangestop + 1)))

    return narr


__all__ = ["Instruction"]



#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from mimiqcircuits import Operation
from mimiqcircuits.operations.operator import AbstractOperator
import mimiqcircuits as mc


class ExpectationValue(Operation):
    r"""Operation to compute and store the expectation value of an Operator in a z-register.

    An expectation value for a pure state :math:`| \psi \rangle` is defined as:

    **Expectation Value for Pure State**

    .. math::
        \langle O \rangle = \langle \psi | O | \psi \rangle

    where :math:`O` is an operator. With respect to a density matrix :math:`\rho`, it's given by:

    **Expectation Value for Density Matrix**

    .. math::
        \langle O \rangle = \mathrm{Tr}(\rho O).

    However, when using quantum trajectories to solve noisy circuits, the expectation
    value is computed with respect to the pure state of each trajectory.

    The argument `op` can be any gate or non-unitary operator.

    .. note::
        ExpectationValue is currently restricted to one and two qubit operators.

    See Also:
        :class:`AbstractOperator`, :class:`Gate`

    Examples:

        In `push!`, the first argument corresponds to the qubit, and the second to the z-register.

        >>> from mimiqcircuits import *
        >>> ExpectationValue(GateX())
        ⟨X⟩

        >>> c = Circuit()
        >>> c.push(ExpectationValue(GateX()), 1, 1)
        2-qubit circuit with 1 instructions:
        └── ⟨X⟩ @ q[1], z[1]
        <BLANKLINE>

        >>> c.push(ExpectationValue(SigmaPlus()), 1, 2)
        2-qubit circuit with 2 instructions:
        ├── ⟨X⟩ @ q[1], z[1]
        └── ⟨SigmaPlus(1)⟩ @ q[1], z[2]
        <BLANKLINE>
    """

    _name = "ExpectationValue"
    _num_zvars = 1
    _num_bits = 0
    _num_zregs = 1
    _num_cregs = 0
    _num_qregs = 1

    def __init__(self, op: AbstractOperator):
        self.op = op

        if not isinstance(op, AbstractOperator):
            raise TypeError(f"cannot get Expectation Value of {op.__class__.__name__}.")

        if not isinstance(op, mc.PauliString) and not (1 <= op._num_qubits <= 2):
            raise ValueError(
                "ExpectationValue only supports 1- or 2-qubit operators unless the operator is a PauliString."
            )

        super().__init__()

        self._num_qubits = self.op._num_qubits
        self._qregsizes = [self._num_qubits]
        self._zregsizes = [1]

    def opname(self):
        return self._name

    @property
    def qregsizes(self):
        return (self._num_qubits,)

    @property
    def cregsizes(self):
        return ()

    @property
    def zregsizes(self):
        return self._zregsizes

    def inverse(self):
        raise NotImplementedError("Cannot inverse an ExpectationValue operation.")

    def power(self, _):
        raise NotImplementedError(
            "Cannot elevate an ExpectationValue operation to any power."
        )

    @staticmethod
    def isunitary():
        return True

    def __str__(self):
        return f"⟨{self.op}⟩"

    def iswrapper(self):
        return False

    def _decompose(self, circ, qubits, bits, zvars):
        return circ.push(self, *qubits, *bits, *zvars)

    def asciiwidth(self, qubits, bits=[], zvars=[]):
        """Calculate the width for ASCII drawing."""
        return len(self.__str__()) + 8

    def get_operation(self):
        return self.op


__all__ = ["ExpectationValue"]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: circuit.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import mimiqcircuits.proto.bitvector_pb2 as bitvector__pb2



DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rcircuit.proto\x1a\x0f\x62itvector.proto\"\x17\n\x06Symbol\x12\r\n\x05value\x18\x01 \x01(\t\"\xa6\x01\n\x03\x41rg\x12#\n\x0e\x61rgvalue_value\x18\x01 \x01(\x0b\x32\t.ArgValueH\x00\x12\x1f\n\x0csymbol_value\x18\x02 \x01(\x0b\x32\x07.SymbolH\x00\x12)\n\x11\x61rgfunction_value\x18\x03 \x01(\x0b\x32\x0c.ArgFunctionH\x00\x12\'\n\x10irrational_value\x18\x04 \x01(\x0e\x32\x0b.IrrationalH\x00\x42\x05\n\x03\x61rg\"^\n\x08\x41rgValue\x12\x17\n\rinteger_value\x18\x01 \x01(\x03H\x00\x12\x16\n\x0c\x64ouble_value\x18\x02 \x01(\x01H\x00\x12\x14\n\nbool_value\x18\x03 \x01(\x08H\x00\x42\x0b\n\targ_value\"?\n\x0b\x41rgFunction\x12\x1c\n\x05mtype\x18\x01 \x01(\x0e\x32\r.FunctionType\x12\x12\n\x04\x61rgs\x18\x02 \x03(\x0b\x32\x04.Arg\"4\n\nComplexArg\x12\x12\n\x04real\x18\x01 \x01(\x0b\x32\x04.Arg\x12\x12\n\x04imag\x18\x02 \x01(\x0b\x32\x04.Arg\"$\n\x08Rational\x12\x0b\n\x03num\x18\x01 \x01(\x03\x12\x0b\n\x03\x64\x65n\x18\x02 \x01(\x03\"@\n\nSimpleGate\x12\x18\n\x05mtype\x18\x01 \x01(\x0e\x32\t.GateType\x12\x18\n\nparameters\x18\x02 \x03(\x0b\x32\x04.Arg\"<\n\nCustomGate\x12\x11\n\tnumqubits\x18\x01 \x01(\x03\x12\x1b\n\x06matrix\x18\x02 \x03(\x0b\x32\x0b.ComplexArg\"B\n\x0bGeneralized\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\x04\x61rgs\x18\x02 \x03(\x0b\x32\x04.Arg\x12\x11\n\tqregsizes\x18\x03 \x03(\x03\"8\n\x07\x43ontrol\x12\x18\n\toperation\x18\x01 \x01(\x0b\x32\x05.Gate\x12\x13\n\x0bnumcontrols\x18\x02 \x01(\x03\"v\n\x05Power\x12\x18\n\toperation\x18\x01 \x01(\x0b\x32\x05.Gate\x12\x14\n\ndouble_val\x18\x02 \x01(\x01H\x00\x12!\n\x0crational_val\x18\x03 \x01(\x0b\x32\t.RationalH\x00\x12\x11\n\x07int_val\x18\x04 \x01(\x03H\x00\x42\x07\n\x05power\"#\n\x07Inverse\x12\x18\n\toperation\x18\x01 \x01(\x0b\x32\x05.Gate\"8\n\x08Parallel\x12\x18\n\toperation\x18\x01 \x01(\x0b\x32\x05.Gate\x12\x12\n\nnumrepeats\x18\x02 \x01(\x03\"7\n\x08GateCall\x12\x17\n\x04\x64\x65\x63l\x18\x01 \x01(\x0b\x32\t.GateDecl\x12\x12\n\x04\x61rgs\x18\x02 \x03(\x0b\x32\x04.Arg\"S\n\x08GateDecl\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\x04\x61rgs\x18\x02 \x03(\x0b\x32\x07.Symbol\x12\"\n\x0cinstructions\x18\x03 \x03(\x0b\x32\x0c.Instruction\"/\n\x0bPauliString\x12\x11\n\tnumqubits\x18\x01 \x01(\x03\x12\r\n\x05pauli\x18\x02 \x01(\t\"\xaf\x02\n\x04Gate\x12!\n\nsimplegate\x18\x01 \x01(\x0b\x32\x0b.SimpleGateH\x00\x12!\n\ncustomgate\x18\x02 \x01(\x0b\x32\x0b.CustomGateH\x00\x12#\n\x0bgeneralized\x18\x03 \x01(\x0b\x32\x0c.GeneralizedH\x00\x12\x1b\n\x07\x63ontrol\x18\x04 \x01(\x0b\x32\x08.ControlH\x00\x12\x17\n\x05power\x18\x05 \x01(\x0b\x32\x06.PowerH\x00\x12\x1b\n\x07inverse\x18\x06 \x01(\x0b\x32\x08.InverseH\x00\x12\x1d\n\x08parallel\x18\x07 \x01(\x0b\x32\t.ParallelH\x00\x12\x1d\n\x08gatecall\x18\x08 \x01(\x0b\x32\t.GateCallH\x00\x12#\n\x0bpaulistring\x18\t \x01(\x0b\x32\x0c.PauliStringH\x00\x42\x06\n\x04gate\"H\n\x0eSimpleOperator\x12\x1c\n\x05mtype\x18\x01 \x01(\x0e\x32\r.OperatorType\x12\x18\n\nparameters\x18\x02 \x03(\x0b\x32\x04.Arg\"@\n\x0e\x43ustomOperator\x12\x11\n\tnumqubits\x18\x01 \x01(\x03\x12\x1b\n\x06matrix\x18\x02 \x03(\x0b\x32\x0b.ComplexArg\"=\n\x0cRescaledGate\x12\x18\n\toperation\x18\x01 \x01(\x0b\x32\x05.Gate\x12\x13\n\x05scale\x18\x02 \x01(\x0b\x32\x04.Arg\"\xb4\x03\n\x08Operator\x12!\n\nsimplegate\x18\x01 \x01(\x0b\x32\x0b.SimpleGateH\x00\x12!\n\ncustomgate\x18\x02 \x01(\x0b\x32\x0b.CustomGateH\x00\x12#\n\x0bgeneralized\x18\x03 \x01(\x0b\x32\x0c.GeneralizedH\x00\x12\x1b\n\x07\x63ontrol\x18\x04 \x01(\x0b\x32\x08.ControlH\x00\x12\x17\n\x05power\x18\x05 \x01(\x0b\x32\x06.PowerH\x00\x12\x1b\n\x07inverse\x18\x06 \x01(\x0b\x32\x08.InverseH\x00\x12\x1d\n\x08parallel\x18\x07 \x01(\x0b\x32\t.ParallelH\x00\x12\x1d\n\x08gatecall\x18\x08 \x01(\x0b\x32\t.GateCallH\x00\x12#\n\x0bpaulistring\x18\t \x01(\x0b\x32\x0c.PauliStringH\x00\x12)\n\x0esimpleoperator\x18\n \x01(\x0b\x32\x0f.SimpleOperatorH\x00\x12)\n\x0e\x63ustomoperator\x18\x0b \x01(\x0b\x32\x0f.CustomOperatorH\x00\x12%\n\x0crescaledgate\x18\x0c \x01(\x0b\x32\r.RescaledGateH\x00\x42\n\n\x08operator\"P\n\x12SimpleKrausChannel\x12 \n\x05mtype\x18\x01 \x01(\x0e\x32\x11.KrausChannelType\x12\x18\n\nparameters\x18\x02 \x03(\x0b\x32\x04.Arg\"E\n\x12\x43ustomKrausChannel\x12\x11\n\tnumqubits\x18\x01 \x01(\x03\x12\x1c\n\toperators\x18\x02 \x03(\x0b\x32\t.Operator\"C\n\x13\x44\x65polarizingChannel\x12\x11\n\tnumqubits\x18\x01 \x01(\x03\x12\x19\n\x0bprobability\x18\x02 \x01(\x0b\x32\x04.Arg\"7\n\x13MixedUnitaryChannel\x12 \n\toperators\x18\x01 \x03(\x0b\x32\r.RescaledGate\"O\n\x0cPauliChannel\x12\x1b\n\rprobabilities\x18\x01 \x03(\x0b\x32\x04.Arg\x12\"\n\x0cpaulistrings\x18\x02 \x03(\x0b\x32\x0c.PauliString\"\x95\x02\n\x0cKrausChannel\x12\x31\n\x12simplekrauschannel\x18\x01 \x01(\x0b\x32\x13.SimpleKrausChannelH\x00\x12\x31\n\x12\x63ustomkrauschannel\x18\x02 \x01(\x0b\x32\x13.CustomKrausChannelH\x00\x12\x33\n\x13\x64\x65polarizingchannel\x18\x03 \x01(\x0b\x32\x14.DepolarizingChannelH\x00\x12\x33\n\x13mixedunitarychannel\x18\x04 \x01(\x0b\x32\x14.MixedUnitaryChannelH\x00\x12%\n\x0cpaulichannel\x18\x05 \x01(\x0b\x32\r.PauliChannelH\x00\x42\x0e\n\x0ckrauschannel\"J\n\x0fSimpleOperation\x12\x1d\n\x05mtype\x18\x01 \x01(\x0e\x32\x0e.OperationType\x12\x18\n\nparameters\x18\x02 \x03(\x0b\x32\x04.Arg\"K\n\x0bIfStatement\x12\x1d\n\toperation\x18\x01 \x01(\x0b\x32\n.Operation\x12\x1d\n\tbitstring\x18\x02 \x01(\x0b\x32\n.BitVector\"\x1c\n\x07\x42\x61rrier\x12\x11\n\tnumqubits\x18\x01 \x01(\x03\"#\n\tAmplitude\x12\x16\n\x02\x62s\x18\x01 \x01(\x0b\x32\n.BitVector\"/\n\x10\x45xpectationValue\x12\x1b\n\x08operator\x18\x01 \x01(\x0b\x32\t.Operator\"9\n\x04Note\x12\x12\n\x08int_note\x18\x01 \x01(\x03H\x00\x12\x15\n\x0b\x64ouble_note\x18\x02 \x01(\x01H\x00\x42\x06\n\x04note\"H\n\x10SimpleAnnotation\x12\x1e\n\x05mtype\x18\x01 \x01(\x0e\x32\x0f.AnnotationType\x12\x14\n\x05notes\x18\x02 \x03(\x0b\x32\x05.Note\"3\n\x08\x44\x65tector\x12\x11\n\tnumqubits\x18\x01 \x01(\x03\x12\x14\n\x05notes\x18\x02 \x03(\x0b\x32\x05.Note\":\n\x11ObservableInclude\x12\x0f\n\x07numbits\x18\x01 \x01(\x03\x12\x14\n\x05notes\x18\x02 \x03(\x0b\x32\x05.Note\"\xe7\x07\n\tOperation\x12!\n\nsimplegate\x18\x01 \x01(\x0b\x32\x0b.SimpleGateH\x00\x12!\n\ncustomgate\x18\x02 \x01(\x0b\x32\x0b.CustomGateH\x00\x12#\n\x0bgeneralized\x18\x03 \x01(\x0b\x32\x0c.GeneralizedH\x00\x12\x1b\n\x07\x63ontrol\x18\x04 \x01(\x0b\x32\x08.ControlH\x00\x12\x17\n\x05power\x18\x05 \x01(\x0b\x32\x06.PowerH\x00\x12\x1b\n\x07inverse\x18\x06 \x01(\x0b\x32\x08.InverseH\x00\x12\x1d\n\x08parallel\x18\x07 \x01(\x0b\x32\t.ParallelH\x00\x12\x1d\n\x08gatecall\x18\x08 \x01(\x0b\x32\t.GateCallH\x00\x12#\n\x0bpaulistring\x18\t \x01(\x0b\x32\x0c.PauliStringH\x00\x12\x31\n\x12simplekrauschannel\x18\n \x01(\x0b\x32\x13.SimpleKrausChannelH\x00\x12\x31\n\x12\x63ustomkrauschannel\x18\x0b \x01(\x0b\x32\x13.CustomKrausChannelH\x00\x12\x33\n\x13\x64\x65polarizingchannel\x18\x0c \x01(\x0b\x32\x14.DepolarizingChannelH\x00\x12\x33\n\x13mixedunitarychannel\x18\r \x01(\x0b\x32\x14.MixedUnitaryChannelH\x00\x12%\n\x0cpaulichannel\x18\x0e \x01(\x0b\x32\r.PauliChannelH\x00\x12)\n\x0esimpleoperator\x18\x0f \x01(\x0b\x32\x0f.SimpleOperatorH\x00\x12)\n\x0e\x63ustomoperator\x18\x10 \x01(\x0b\x32\x0f.CustomOperatorH\x00\x12%\n\x0crescaledgate\x18\x11 \x01(\x0b\x32\r.RescaledGateH\x00\x12+\n\x0fsimpleoperation\x18\x12 \x01(\x0b\x32\x10.SimpleOperationH\x00\x12#\n\x0bifstatement\x18\x13 \x01(\x0b\x32\x0c.IfStatementH\x00\x12\x1b\n\x07\x62\x61rrier\x18\x14 \x01(\x0b\x32\x08.BarrierH\x00\x12\x1f\n\tamplitude\x18\x15 \x01(\x0b\x32\n.AmplitudeH\x00\x12-\n\x10\x65xpectationvalue\x18\x16 \x01(\x0b\x32\x11.ExpectationValueH\x00\x12\x1d\n\x08\x64\x65tector\x18\x17 \x01(\x0b\x32\t.DetectorH\x00\x12+\n\robservableinc\x18\x18 \x01(\x0b\x32\x12.ObservableIncludeH\x00\x12-\n\x10simpleannotation\x18\x19 \x01(\x0b\x32\x11.SimpleAnnotationH\x00\x42\x0b\n\toperation\"b\n\x0bInstruction\x12\x1d\n\toperation\x18\x01 \x01(\x0b\x32\n.Operation\x12\x10\n\x08qtargets\x18\x02 \x03(\x03\x12\x10\n\x08\x63targets\x18\x03 \x03(\x03\x12\x10\n\x08ztargets\x18\x04 \x03(\x03\"-\n\x07\x43ircuit\x12\"\n\x0cinstructions\x18\x01 \x03(\x0b\x32\x0c.Instruction*\x1f\n\nIrrational\x12\x06\n\x02PI\x10\x00\x12\t\n\x05\x45ULER\x10\x01*m\n\x0c\x46unctionType\x12\x07\n\x03\x41\x44\x44\x10\x00\x12\x07\n\x03MUL\x10\x01\x12\x07\n\x03\x44IV\x10\x02\x12\x07\n\x03POW\x10\x03\x12\x07\n\x03SIN\x10\x05\x12\x07\n\x03\x43OS\x10\x06\x12\x07\n\x03TAN\x10\x07\x12\x07\n\x03\x45XP\x10\x08\x12\x07\n\x03LOG\x10\t\x12\x0c\n\x08IDENTITY\x10\n*\xf6\x02\n\x08GateType\x12\n\n\x06GateID\x10\x00\x12\t\n\x05GateX\x10\x01\x12\t\n\x05GateY\x10\x02\x12\t\n\x05GateZ\x10\x03\x12\t\n\x05GateH\x10\x04\x12\x0b\n\x07GateHXY\x10\x05\x12\x0b\n\x07GateHYZ\x10\x06\x12\t\n\x05GateS\x10\x07\x12\t\n\x05GateT\x10\x08\x12\t\n\x05\x44\x65lay\x10\t\x12\t\n\x05GateU\x10\n\x12\t\n\x05GateP\x10\x0b\x12\n\n\x06GateRX\x10\x0c\x12\n\n\x06GateRY\x10\r\x12\n\n\x06GateRZ\x10\x0e\x12\t\n\x05GateR\x10\x0f\x12\n\n\x06GateU1\x10\x10\x12\n\n\x06GateU2\x10\x11\x12\n\n\x06GateU3\x10\x12\x12\x0c\n\x08GateSWAP\x10\x13\x12\r\n\tGateISWAP\x10\x14\x12\x0b\n\x07GateECR\x10\x15\x12\x0b\n\x07GateDCX\x10\x16\x12\x0b\n\x07GateRXX\x10\x17\x12\x0b\n\x07GateRYY\x10\x18\x12\x0b\n\x07GateRZZ\x10\x19\x12\x0b\n\x07GateRZX\x10\x1a\x12\x10\n\x0cGateXXplusYY\x10\x1b\x12\x11\n\rGateXXminusYY\x10\x1c*\xbf\x02\n\x0cOperatorType\x12\x0e\n\nSigmaMinus\x10\x00\x12\r\n\tSigmaPlus\x10\x01\x12\r\n\tSigma0001\x10\x02\x12\r\n\tSigma0010\x10\x03\x12\r\n\tSigma0011\x10\x04\x12\r\n\tSigma0110\x10\x05\x12\r\n\tSigma0111\x10\x06\x12\r\n\tSigma1011\x10\x07\x12\x0e\n\nProjector0\x10\x08\x12\x0e\n\nProjector1\x10\t\x12\x0f\n\x0bProjector00\x10\n\x12\x0f\n\x0bProjector01\x10\x0b\x12\x0f\n\x0bProjector10\x10\x0c\x12\x0f\n\x0bProjector11\x10\r\x12\x0f\n\x0bProjectorX0\x10\x0e\x12\x0f\n\x0bProjectorX1\x10\x0f\x12\x0f\n\x0bProjectorY0\x10\x10\x12\x0f\n\x0bProjectorY1\x10\x11\x12\x0e\n\nDiagonalOp\x10\x12*\x80\x02\n\x10KrausChannelType\x12\n\n\x06ResetX\x10\x00\x12\n\n\x06ResetY\x10\x01\x12\n\n\x06ResetZ\x10\x02\x12\x14\n\x10\x41mplitudeDamping\x10\x03\x12\x1f\n\x1bGeneralizedAmplitudeDamping\x10\x04\x12\x19\n\x15PhaseAmplitudeDamping\x10\x05\x12\x10\n\x0cThermalNoise\x10\x06\x12\n\n\x06PauliX\x10\x07\x12\n\n\x06PauliY\x10\x08\x12\n\n\x06PauliZ\x10\t\x12\x14\n\x10ProjectiveNoiseX\x10\n\x12\x14\n\x10ProjectiveNoiseY\x10\x0b\x12\x14\n\x10ProjectiveNoiseZ\x10\x0c*\xdd\x01\n\rOperationType\x12\x0c\n\x08MeasureX\x10\x00\x12\x0c\n\x08MeasureY\x10\x01\x12\x0c\n\x08MeasureZ\x10\x02\x12\r\n\tMeasureXX\x10\x03\x12\r\n\tMeasureYY\x10\x04\x12\r\n\tMeasureZZ\x10\x05\x12\x11\n\rMeasureResetX\x10\x06\x12\x11\n\rMeasureResetY\x10\x07\x12\x11\n\rMeasureResetZ\x10\x08\x12\x0b\n\x07\x42ondDim\x10\t\x12\x0f\n\x0bSchmidtRank\x10\n\x12\x15\n\x11VonNeumannEntropy\x10\x0b\x12\x07\n\x03Not\x10\x0c*F\n\x0e\x41nnotationType\x12\x14\n\x10QubitCoordinates\x10\x00\x12\x14\n\x10ShiftCoordinates\x10\x01\x12\x08\n\x04Tick\x10\x03\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'circuit_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _IRRATIONAL._serialized_start=4393
  _IRRATIONAL._serialized_end=4424
  _FUNCTIONTYPE._serialized_start=4426
  _FUNCTIONTYPE._serialized_end=4535
  _GATETYPE._serialized_start=4538
  _GATETYPE._serialized_end=4912
  _OPERATORTYPE._serialized_start=4915
  _OPERATORTYPE._serialized_end=5234
  _KRAUSCHANNELTYPE._serialized_start=5237
  _KRAUSCHANNELTYPE._serialized_end=5493
  _OPERATIONTYPE._serialized_start=5496
  _OPERATIONTYPE._serialized_end=5717
  _ANNOTATIONTYPE._serialized_start=5719
  _ANNOTATIONTYPE._serialized_end=5789
  _SYMBOL._serialized_start=34
  _SYMBOL._serialized_end=57
  _ARG._serialized_start=60
  _ARG._serialized_end=226
  _ARGVALUE._serialized_start=228
  _ARGVALUE._serialized_end=322
  _ARGFUNCTION._serialized_start=324
  _ARGFUNCTION._serialized_end=387
  _COMPLEXARG._serialized_start=389
  _COMPLEXARG._serialized_end=441
  _RATIONAL._serialized_start=443
  _RATIONAL._serialized_end=479
  _SIMPLEGATE._serialized_start=481
  _SIMPLEGATE._serialized_end=545
  _CUSTOMGATE._serialized_start=547
  _CUSTOMGATE._serialized_end=607
  _GENERALIZED._serialized_start=609
  _GENERALIZED._serialized_end=675
  _CONTROL._serialized_start=677
  _CONTROL._serialized_end=733
  _POWER._serialized_start=735
  _POWER._serialized_end=853
  _INVERSE._serialized_start=855
  _INVERSE._serialized_end=890
  _PARALLEL._serialized_start=892
  _PARALLEL._serialized_end=948
  _GATECALL._serialized_start=950
  _GATECALL._serialized_end=1005
  _GATEDECL._serialized_start=1007
  _GATEDECL._serialized_end=1090
  _PAULISTRING._serialized_start=1092
  _PAULISTRING._serialized_end=1139
  _GATE._serialized_start=1142
  _GATE._serialized_end=1445
  _SIMPLEOPERATOR._serialized_start=1447
  _SIMPLEOPERATOR._serialized_end=1519
  _CUSTOMOPERATOR._serialized_start=1521
  _CUSTOMOPERATOR._serialized_end=1585
  _RESCALEDGATE._serialized_start=1587
  _RESCALEDGATE._serialized_end=1648
  _OPERATOR._serialized_start=1651
  _OPERATOR._serialized_end=2087
  _SIMPLEKRAUSCHANNEL._serialized_start=2089
  _SIMPLEKRAUSCHANNEL._serialized_end=2169
  _CUSTOMKRAUSCHANNEL._serialized_start=2171
  _CUSTOMKRAUSCHANNEL._serialized_end=2240
  _DEPOLARIZINGCHANNEL._serialized_start=2242
  _DEPOLARIZINGCHANNEL._serialized_end=2309
  _MIXEDUNITARYCHANNEL._serialized_start=2311
  _MIXEDUNITARYCHANNEL._serialized_end=2366
  _PAULICHANNEL._serialized_start=2368
  _PAULICHANNEL._serialized_end=2447
  _KRAUSCHANNEL._serialized_start=2450
  _KRAUSCHANNEL._serialized_end=2727
  _SIMPLEOPERATION._serialized_start=2729
  _SIMPLEOPERATION._serialized_end=2803
  _IFSTATEMENT._serialized_start=2805
  _IFSTATEMENT._serialized_end=2880
  _BARRIER._serialized_start=2882
  _BARRIER._serialized_end=2910
  _AMPLITUDE._serialized_start=2912
  _AMPLITUDE._serialized_end=2947
  _EXPECTATIONVALUE._serialized_start=2949
  _EXPECTATIONVALUE._serialized_end=2996
  _NOTE._serialized_start=2998
  _NOTE._serialized_end=3055
  _SIMPLEANNOTATION._serialized_start=3057
  _SIMPLEANNOTATION._serialized_end=3129
  _DETECTOR._serialized_start=3131
  _DETECTOR._serialized_end=3182
  _OBSERVABLEINCLUDE._serialized_start=3184
  _OBSERVABLEINCLUDE._serialized_end=3242
  _OPERATION._serialized_start=3245
  _OPERATION._serialized_end=4244
  _INSTRUCTION._serialized_start=4246
  _INSTRUCTION._serialized_end=4344
  _CIRCUIT._serialized_start=4346
  _CIRCUIT._serialized_end=4391
# @@protoc_insertion_point(module_scope)



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.power import Power
from mimiqcircuits.operations.inverse import Inverse
from mimiqcircuits.operations.gates.standard.pauli import GateZ
from mimiqcircuits.operations.gates.standard.u import GateU
from mimiqcircuits.operations.utils import control_one_defined
import mimiqcircuits as mc
from symengine import pi


class GateS(Power):
    r"""Single qubit gate S.

    It induces a :math:`\frac{\pi}{2}` phase gate.

    **Matrix representation:**

    .. math::
        \operatorname{S} = \begin{pmatrix}
            1 & 0 \\
            0 & i
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateS()
        S
        >>> GateS().matrix()
        [1.0, 0]
        [0, 0.0 + 1.0*I]
        <BLANKLINE>
        >>> c = Circuit().push(GateS(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── S @ q[0]
        <BLANKLINE>
        >>> GateS().power(2), GateS().inverse()
        (Z, S†)
        >>> GateS().decompose()
        1-qubit circuit with 1 instructions:
        └── U(0, 0, (1/2)*pi, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "S"

    def __init__(self):
        super().__init__(GateZ(), 1 / 2)

    def isopalias(self):
        return True

    def inverse(self):
        return GateSDG()

    def _power(self, p):
        pmod = p % 4

        # sqrt(S) = T
        if pmod == 1 / 2:
            return mc.GateT()

        # T * T^7 = S^2 * S^2 = Z * Z = ID => T^7 = TDG
        if pmod == 7 / 2:
            return mc.GateTDG()

        # Z^(2n) = ID => S^(4n) = ID
        if p % 4 == 0:
            return mc.GateID()

        # Z^(2n + 1) = Z => S^(4n + 1) = S
        if p % 4 == 1:
            return mc.GateS()

        #  S^(4n + 3) = SDG
        if p % 4 == 3:
            return mc.GateSDG()

        # sqrt(Z) = S
        if p % 2 == 0:
            return mc.GateZ().power(p / 2)

    def _control(self, n):
        return control_one_defined(n, self, mc.GateCS())

    def __str__(self):
        return f"{self.name}"

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(0, 0, pi / 2), q)
        return circ


class GateSDG(Inverse):
    r"""Single qubit S-dagger gate (conjugate transpose of the S gate).

    **Matrix representation:**

    .. math::
        \operatorname{S}^\dagger = \begin{pmatrix}
            1 & 0 \\
            0 & -i
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateSDG()
        S†
        >>> GateSDG().matrix()
        [1.0, 0]
        [0, 6.12323399573677e-17 - 1.0*I]
        <BLANKLINE>
        >>> c = Circuit().push(GateSDG(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── S† @ q[0]
        <BLANKLINE>
        >>> GateSDG().power(2), GateSDG().inverse()
        (S†**2, S)
        >>> GateSDG().decompose()
        1-qubit circuit with 1 instructions:
        └── U(0, 0, (-1/2)*pi, 0.0) @ q[0]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(GateS())

    def isopalias(self):
        return True

    def inverse(self):
        return GateS()

    def _control(self, n):
        return control_one_defined(n, self, mc.GateCSDG())

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(0, 0, -pi / 2), q)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits import Operation

class Not(Operation):
    r"""
    Not operation.

    Represents a NOT operation that can be added to quantum circuits.
    This operation  inverts a classical bit.

    Examples:

        >>> from mimiqcircuits import *
        >>> not_op = Not()
        >>> not_op.name
        '!'
        >>> c = Circuit()
        >>> c.push(Not(), 1)
        0-qubit circuit with 1 instructions:
        └── ! @ c[1]
        <BLANKLINE>
    """

    _name = "!"
    _num_bits = 1
    _num_qubits = 0
    _num_cregs = 0
    _num_zvars = 0
    _qregsizes = []
    _cregsizes = [1]

    def inverse(self):
        return Not()

    def iswrapper(self):
        return False

    def get_operation(self):
        return self

    def __repr__(self):
        return f"{self._name}"
    
__all__ = ["Not"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import symengine as se
from typing import TypeVar, Generic
from mimiqcircuits.operations.operator import AbstractOperator
from mimiqcircuits.operations.gates.gate import Gate
import sympy as sp

T = TypeVar("T", bound=Gate)  # T is a type variable bound to Gate


class RescaledGate(AbstractOperator, Generic[T]):
    r"""RescaledGate operation.

    The `RescaledGate` represents an operation where a quantum gate is rescaled by a factor `p`, 
    typically between 0 and 1. This rescaling modifies the action of the gate, multiplying its 
    matrix representation by `p`. The gate to be rescaled must be an instance of the `Gate` class, 
    and `p` must be a valid scalar (real or symbolic) in the range [0, 1].

    Args:
        - gate (Gate): The quantum gate to be rescaled.
        - p (float or symbolic): The rescaling factor, must be between 0 and 1.

    Raises:
        - TypeError: If the provided gate is not an instance of `Gate`.
        - ValueError: If the rescaling factor `p` is not between 0 and 1.

    Methods:
        - get_operation(): Returns the underlying gate.
        - get_param(name): Retrieves the value of a parameter (such as `p`).
        - _matrix(*args): Returns the matrix representation of the rescaled gate.
        - get_scale(): Returns the scaling factor `p`.
        - rescale(p): Returns a new `RescaledGate` with a different scaling factor.
        - rescale_in_place(p): Rescales the gate in place by multiplying the current `p` with the new value.
        - evaluate(d): Substitutes values in the parameters (useful when symbolic values are used).

    Examples:

        Creating and rescaling a gate:

        >>> from mimiqcircuits import *
        >>> g = GateH()
        >>> rg = RescaledGate(g, 0.5)

        >>> rg.get_scale()
        0.5

        >>> rg.rescale(0.8)
        0.4*H
    """

    _name = "RescaledGate"

    def __init__(self, gate: T, p):
        if not isinstance(gate, Gate):
            raise TypeError(
                f"Expected gate to be an instance of Gate, got {type(gate).__name__}"
            )

        super().__init__()
        self._num_bits = gate.num_bits
        self._num_qubits = gate.num_qubits
        self._num_zvars = gate.num_zvars
        self._num_cregs = gate._num_cregs
        self._num_qregs = gate._num_qregs
        self._num_zregs = gate._num_zregs
        self._cregsizes = gate._cregsizes
        self._qregsizes = gate._qregsizes
        self._zregsizes = gate._zregsizes

        if not isinstance(p, (se.Basic, sp.Basic)) and (float(p) < 0 or float(p) > 1):
            raise ValueError("Value of p must be between 0 and 1.")
        self.gate = gate
        self.p = p

    def __repr__(self):
        return self.__str__()

    def get_operation(self):
        return self.gate

    def get_param(self, name):
        if name == "p":
            return self.p
        return self.gate.get_param(name)

    def _matrix(self, *args):
        return self.p * self.gate._matrix(*args)

    def get_scale(self):
        return self.p

    def rescale(self, p):
        return RescaledGate(self.gate, self.p * p)

    def rescale_in_place(self, p):
        self.p *= p
        return self

    def evaluate(self, d):
        return RescaledGate(self.gate.evaluate(d), evaluate(self.p, d))

    def __str__(self):
        mult = "*" if getattr(self, "compact", True) else " * "
        return f"{self.p}{mult}{self.gate}"


def evaluate(param, d):
    # Replace symbols in param according to dictionary d, this is just a placeholder
    return param.subs(d) if isinstance(param, (se.Basic, sp.Basic)) else param



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc


class LazyArg:
    def __str__(self):
        return "?"

    def __repr__(self):
        return self.__str__()


class LazyExpr:
    def __init__(self, obj, *args):
        self.obj = obj
        self.args = list(args)

    def _lazy_recursive_evaluate(self, args):
        actual = []
        for arg in self.args:
            if isinstance(arg, LazyArg):
                new_arg = args.pop()
            elif isinstance(arg, LazyExpr):
                new_arg = arg._lazy_recursive_evaluate(args)
            else:
                new_arg = arg
            actual.append(new_arg)
        return self.obj(*actual)

    def __call__(self, *args):
        return self._lazy_recursive_evaluate(list(reversed(args)))

    def inverse(self):
        return LazyExpr(inverse, self)

    def parallel(self, *args):
        if len(args) == 0:
            return LazyExpr(parallel, LazyArg(), self)
        elif len(args) == 1:
            num_repeats = args[0]
            return LazyExpr(parallel, num_repeats, self)
        else:
            raise TypeError("Invalid number of arguments.")

    def control(self, *args):
        if len(args) == 0:
            return LazyExpr(control, LazyArg(), self)
        elif len(args) == 1:
            num_controls = args[0]
            return LazyExpr(control, num_controls, self)
        else:
            raise TypeError("Invalid number of arguments.")

    def power(self, *args):
        if len(args) == 0:
            return LazyExpr(power, self, LazyArg())
        elif len(args) == 1:
            p = args[0]
            return LazyExpr(power, self, p)
        else:
            return TypeError("Invalid number of arguments.")

    def __str__(self):
        return f"lazy {self.obj.__name__}({', '.join(map(str, self.args))})"

    def __repr__(self):
        return str(self)


def control(*args):
    if len(args) == 1:
        arg = args[0]
        if isinstance(arg, mc.Operation):
            return LazyExpr(control, LazyArg(), arg)
        elif isinstance(arg, int):
            return LazyExpr(control, arg, LazyArg())
        else:
            raise TypeError("Invalid argument type.")
    elif len(args) == 2:
        numcontrols, gate = args
        return gate.control(numcontrols)
    else:
        raise TypeError("Invalid number of arguments.")


def parallel(*args):
    if len(args) == 1:
        arg = args[0]
        if isinstance(arg, mc.Operation):
            return LazyExpr(parallel, LazyArg(), arg)
        elif isinstance(arg, int):
            return LazyExpr(parallel, arg, LazyArg())
        else:
            raise TypeError("Invalid argument type.")
    elif len(args) == 2:
        num_repeats, gate = args
        return gate.parallel(num_repeats)
    else:
        raise TypeError("Invalid number of arguments.")


def inverse(op):
    return op.inverse()


def power(*args):
    if len(args) == 1:
        arg = args[0]
        if isinstance(arg, mc.Operation):
            return LazyExpr(power, arg, LazyArg())
        elif isinstance(arg, int):
            return LazyExpr(power, LazyArg(), arg)
        else:
            raise TypeError("Invalid argument type.")
    elif len(args) == 2:
        gate, p = args
        return gate.power(p)
    else:
        raise TypeError("Invalid number of arguments.")


__all__ = ["LazyArg", "LazyExpr", "control", "parallel", "inverse", "power"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.gates.gate as mcg
from mimiqcircuits.operations.gates.standard.phase import GateP
import mimiqcircuits.lazy as lz
from symengine import pi


class PhaseGradient(mcg.Gate):
    """
    Phase Gradient gate

    A phase gradient gate applies a phase shift to a quantum register of `n` qubits,
    where each computational basis state `|k⟩` experiences a phase
    proportional to its integer value `k`

    Args:
        n (int): The number of qubits in the quantum register.

    Returns:
        PhaseGradient: The PhaseGradient gate.

    Attributes:
        name (str): The name of the operation.
        num_qubits (int): The number of qubits in the quantum register.
        qregsizes (list of int): The sizes of the quantum registers.

    Examples:
        >>> from mimiqcircuits import *
        >>> c=Circuit()
        >>> c.push(PhaseGradient(2),9,8)
        10-qubit circuit with 1 instructions:
        └── PhaseGradient @ q[9,8]
        <BLANKLINE>

    """

    _name = "PhaseGradient"
    _num_qregs = 1

    def __init__(self, num_qubits):
        super().__init__()

        self._num_qubits = num_qubits
        self._qregsizes = [
            num_qubits,
        ]
        self._params = [num_qubits]

        self._num_bits = 0

    def __new__(cls, *args):
        if len(args) == 0:
            return lz.LazyExpr(PhaseGradient, lz.LazyArg())
        elif len(args) == 1:
            return object.__new__(cls)
        else:
            raise ValueError("Invalid number of arguments.")

    def _decompose(self, circ, qubits, bits, _):
        for i in range(len(qubits) - 1, -1, -1):
            phase = 1 * pi / 2.0 ** (i)
            circ.push(GateP(phase), qubits[i])
        return circ

    def _matrix(self):
        raise NotImplementedError(
            "Matrix representation for PhaseGradient is not implemented."
        )


__all__ = ["PhaseGradient"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.gates.gate as mcg
from mimiqcircuits.operations.gates.standard.cphase import GateCP
from mimiqcircuits.operations.gates.standard.hadamard import GateH
import mimiqcircuits.lazy as lz
from symengine import pi


class QFT(mcg.Gate):
    """
    Quantum Fourier transform.

    Performs the quantum Fourier transform on a register of `n` qubits.

    Args:
        n (int): The number of qubits in the quantum register.

    Raises:
        ValueError: If the number of qubits is less than 1.

    Returns:
        QFT: The Quantum Fourier Transform operation.

    Attributes:
        name (str): The name of the operation.
        num_qubits (int): The number of qubits in the quantum register.
        qregsizes (list of int): The sizes of the quantum registers.

    Examples:

        >>> from mimiqcircuits import *
        >>> c=Circuit()
        >>> c.push(QFT(2),1,2)
        3-qubit circuit with 1 instructions:
        └── QFT @ q[1,2]
        <BLANKLINE>

    """

    _name = "QFT"
    _num_qregs = 1
    _num_qubits = None

    def __init__(self, num_qubits):
        if num_qubits < 1:
            raise ValueError("Number of qubits must be greater than 0")

        super().__init__()

        self._num_qubits = num_qubits
        self._qregsizes = [num_qubits]
        self._params = [num_qubits]

    def __new__(cls, *args):
        if len(args) == 0:
            return lz.LazyExpr(QFT, lz.LazyArg())
        elif len(args) == 1:
            return object.__new__(cls)
        else:
            raise ValueError("Invalid number of arguments.")

    def _matrix(self):
        raise NotImplementedError(
            "Matrix representation for Quantum Fourier Transform is not implemented."
        )

    def _decompose(self, circ, qubits, bits, _):
        q = qubits[::-1]

        circ.push(GateH(), q[0])

        for i in range(1, self.num_qubits):
            for j in range(i):
                angle = pi / (2.0 ** (i - j))
                circ.push(GateCP(angle), q[i], q[j])

            circ.push(GateH(), q[i])

        return circ


__all__ = ["QFT"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import (
    annotations,
)  # Required for postponed evaluation of type hints in Python 3.7+
import mimiqcircuits as mc
from mimiqcircuits.instruction import Instruction
import copy
from collections.abc import Iterable
from itertools import repeat
import shutil
from typing import List, Union
import random
import numpy as np


class Circuit:
    """Representation of a quantum circuit.

    Operation can be added one by one to a circuit with the
    ``c.push(operation, targets...)`` function

    Args:
        instructions (list of Instruction): Instructiuons to add at
            construction.

    Raises:
        TypeError: If initialization list contains non-Instruction objects.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import pi

        Create a new circuit object

        >>> c = Circuit()

        Add a GateX (Pauli-X) gate on qubit 0

        >>> c.push(GateX(), 0)
        1-qubit circuit with 1 instructions:
        └── X @ q[0]
        <BLANKLINE>

        Add a Controlled-NOT (CX) gate with control qubit 0 and target qubit 1

        >>> c.push(GateCX(), 0, 1)
        2-qubit circuit with 2 instructions:
        ├── X @ q[0]
        └── CX @ q[0], q[1]
        <BLANKLINE>

        Add a Parametric GateRX gate with parameters pi/4

        >>> c.push(GateRX(pi / 4),0)
        2-qubit circuit with 3 instructions:
        ├── X @ q[0]
        ├── CX @ q[0], q[1]
        └── RX((1/4)*pi) @ q[0]
        <BLANKLINE>

        Add a Reset gate on qubit 0

        >>> c.push(Reset(), 0)
        2-qubit circuit with 4 instructions:
        ├── X @ q[0]
        ├── CX @ q[0], q[1]
        ├── RX((1/4)*pi) @ q[0]
        └── Reset @ q[0]
        <BLANKLINE>

        Add a Barrier gate on qubits 0 and 1

        >>> c.push(Barrier(2), 0, 1)
        2-qubit circuit with 5 instructions:
        ├── X @ q[0]
        ├── CX @ q[0], q[1]
        ├── RX((1/4)*pi) @ q[0]
        ├── Reset @ q[0]
        └── Barrier @ q[0,1]
        <BLANKLINE>

        Add a Measurement gate on qubit 0, storing the result in bit 0.

        >>> c.push(Measure(), 0, 0)
        2-qubit circuit with 6 instructions:
        ├── X @ q[0]
        ├── CX @ q[0], q[1]
        ├── RX((1/4)*pi) @ q[0]
        ├── Reset @ q[0]
        ├── Barrier @ q[0,1]
        └── M @ q[0], c[0]
        <BLANKLINE>

        Add a Control gate with GateX as the target gate. The first 3
        qubits are the control qubits.

        >>> c.push(Control(3, GateX()), 0, 1, 2, 3)
        4-qubit circuit with 7 instructions:
        ├── X @ q[0]
        ├── CX @ q[0], q[1]
        ├── RX((1/4)*pi) @ q[0]
        ├── Reset @ q[0]
        ├── Barrier @ q[0,1]
        ├── M @ q[0], c[0]
        └── C₃X @ q[0,1,2], q[3]
        <BLANKLINE>

        Add a 3-qubit Parallel gate with GateX

        >>> c.push(Parallel(3,GateX()),0, 1, 2)
        4-qubit circuit with 8 instructions:
        ├── X @ q[0]
        ├── CX @ q[0], q[1]
        ├── RX((1/4)*pi) @ q[0]
        ├── Reset @ q[0]
        ├── Barrier @ q[0,1]
        ├── M @ q[0], c[0]
        ├── C₃X @ q[0,1,2], q[3]
        └── ⨷ ³ X @ q[0], q[1], q[2]
        <BLANKLINE>

        To add operations without constructing them first, use the
        `c.emplace(...)` function.

    Available operations
    --------------------

    **Gates**

    **Single qubit gates**
        :func:`GateX` :func:`GateY` :func:`GateZ` :func:`GateH`
        :func:`GateS` :func:`GateSDG`
        :func:`GateT` :func:`GateTDG`
        :func:`GateSX` :func:`GateSXDG`
        :func:`GateID`

    **Single qubit gates (parametric)**
        :func:`GateU` :func:`GateP`
        :func:`GateRX` :func:`GateRY` :func:`GateRZ` :func:`GateP`

    **Two qubit gates**
        :func:`GateCX` :func:`GateCY` :func:`GateCZ`
        :func:`GateCH`
        :func:`GateSWAP` :func:`GateISWAP`
        :func:`GateCS` :func:`GateCSX`
        :func:`GateECR` :func:`GateDCX`

    **Two qubit gates (parametric)**
        :func:`GateCU`
        :func:`GateCP`
        :func:`GateCRX` :func:`GateCRY` :func:`GateCRZ`
        :func:`GateRXX` :func:`GateRYY` :func:`GateRZZ`
        :func:`GateXXplusYY` :func:`GateXXminusYY`

    **Other**
        :func:`GateCustom`

    **No-ops**
        :func:`Barrier`

    **Non-unitary operations**
        :func:`Measure` :func:`Reset`

    **Composite operations**
        :func:`Control` :func:`Parallel`

    **Power & Inverse operations**
        :func:`Power`
        :func:`Inverse`

    **Generalized gates**
        :func:`QFT` :func:`PhaseGradient`
    """

    def __init__(self, instructions=None):
        if instructions is None:
            instructions = []

        if not isinstance(instructions, list):
            raise TypeError(
                "Circuit should be initialized with a list of Instruction")

        for instruction in instructions:
            if not isinstance(instruction, Instruction):
                raise TypeError("Non Gate object passed to constructor.")

        self.instructions = instructions

    def num_qubits(self):
        """
        Returns the number of qubits in the circuit.
        """
        n = -1
        for instruction in self.instructions:
            qubits = instruction.qubits
            if len(qubits) == 0:
                continue
            m = max(qubits)
            if m > n:
                n = m

        return n + 1

    def num_bits(self):
        """
        Returns the number of bits in the circuit.
        """
        n = -1
        for instruction in self.instructions:
            bits = instruction.bits
            if len(bits) == 0:
                continue
            m = max(bits)
            if m > n:
                n = m

        return n + 1

    def num_zvars(self):
        """
        Returns the number of z-variables in the circuit.
        """
        n = -1
        for instruction in self.instructions:
            zbits = instruction.zvars
            if len(zbits) == 0:
                continue
            m = max(zbits)
            if m > n:
                n = m

        return n + 1

    def empty(self):
        """
        Checks if the circuit is empty.
        """
        return len(self.instructions) == 0

    def push(self, operation, *args):
        """
        Adds an Operation or an Instruction to the end of the circuit.

        Args:
            operation (Operation or Instruction): the quantum operation to add.

            args (integers or iterables): Target qubits and bits for the operation
                (not instruction), given as variable number of arguments.

        Raises:
            TypeError: If operation is not an Operation object.

            ValueError: If the number of arguments is incorrect or the
                 target qubits specified are invalid.

        Examples:
            Adding multiple operations to the Circuit (The args can be
            integers or integer-valued iterables)

            >>> from mimiqcircuits import *
            >>> from symengine import pi
            >>> c = Circuit()
            >>> c.push(GateH(), 0)
            1-qubit circuit with 1 instructions:
            └── H @ q[0]
            <BLANKLINE>
            >>> c.push(GateT(), 0)
            1-qubit circuit with 2 instructions:
            ├── H @ q[0]
            └── T @ q[0]
            <BLANKLINE>
            >>> c.push(GateH(), [0,2])
            3-qubit circuit with 4 instructions:
            ├── H @ q[0]
            ├── T @ q[0]
            ├── H @ q[0]
            └── H @ q[2]
            <BLANKLINE>
            >>> c.push(GateS(), 0)
            3-qubit circuit with 5 instructions:
            ├── H @ q[0]
            ├── T @ q[0]
            ├── H @ q[0]
            ├── H @ q[2]
            └── S @ q[0]
            <BLANKLINE>
            >>> c.push(GateCX(), [2, 0], 1)
            3-qubit circuit with 7 instructions:
            ├── H @ q[0]
            ├── T @ q[0]
            ├── H @ q[0]
            ├── H @ q[2]
            ├── S @ q[0]
            ├── CX @ q[2], q[1]
            └── CX @ q[0], q[1]
            <BLANKLINE>
            >>> c.push(GateH(), 0)
            3-qubit circuit with 8 instructions:
            ├── H @ q[0]
            ├── T @ q[0]
            ├── H @ q[0]
            ├── H @ q[2]
            ├── S @ q[0]
            ├── CX @ q[2], q[1]
            ├── CX @ q[0], q[1]
            └── H @ q[0]
            <BLANKLINE>
            >>> c.push(Barrier(3), *range(3)) # equivalent to c.push(Barrier(3), 0, 1, 2)
            3-qubit circuit with 9 instructions:
            ├── H @ q[0]
            ├── T @ q[0]
            ├── H @ q[0]
            ├── H @ q[2]
            ├── S @ q[0]
            ├── CX @ q[2], q[1]
            ├── CX @ q[0], q[1]
            ├── H @ q[0]
            └── Barrier @ q[0,1,2]
            <BLANKLINE>
            >>> c.push(Measure(), range(3), range(3))
            3-qubit circuit with 12 instructions:
            ├── H @ q[0]
            ├── T @ q[0]
            ├── H @ q[0]
            ├── H @ q[2]
            ├── S @ q[0]
            ├── CX @ q[2], q[1]
            ├── CX @ q[0], q[1]
            ├── H @ q[0]
            ├── Barrier @ q[0,1,2]
            ├── M @ q[0], c[0]
            ├── M @ q[1], c[1]
            └── M @ q[2], c[2]
            <BLANKLINE>
            >>> c
            3-qubit circuit with 12 instructions:
            ├── H @ q[0]
            ├── T @ q[0]
            ├── H @ q[0]
            ├── H @ q[2]
            ├── S @ q[0]
            ├── CX @ q[2], q[1]
            ├── CX @ q[0], q[1]
            ├── H @ q[0]
            ├── Barrier @ q[0,1,2]
            ├── M @ q[0], c[0]
            ├── M @ q[1], c[1]
            └── M @ q[2], c[2]
            <BLANKLINE>
        """
        args = tuple(int(arg) if isinstance(
            arg, np.integer) else arg for arg in args)
        N = 0
        M = 0
        Z = 0
        L = len(args)

        if isinstance(operation, Instruction):
            if L != 0:
                raise (
                    ValueError(
                        "No extra arguments allowed when pushing an instruction."
                    )
                )

            self.instructions.append(operation)
            return self

        if not isinstance(operation, mc.Operation):
            raise (TypeError("Non Operation object passed to push."))

        # Check if the operation is an abstract operator but not a gate
        if isinstance(operation, mc.AbstractOperator) and not isinstance(
            operation, mc.Gate
        ):
            raise ValueError(
                f"Cannot add an abstract operator: {type(operation).__name__} that is not a Gate to the circuit."
            )

        N = operation.num_qubits
        M = operation.num_bits
        Z = operation.num_zvars

        if L != N + M + Z:
            raise (
                ValueError(
                    f"Wrong number of target qubits and bits and zvars, given {L} for a {N} qubits + {M} bits operation + {Z} z variables."
                )
            )

        targets = []
        hasiterable = False
        for arg in args[:-1]:
            if isinstance(arg, int):
                targets.append(repeat(arg))
            elif isinstance(arg, Iterable):
                targets.append(arg)
                hasiterable = True
            else:
                raise TypeError(
                    f"Invalid target type for {arg} of type {type(arg)}")
        if args:
            larg = args[-1]
            if isinstance(larg, int):
                if hasiterable:
                    targets.append(repeat(larg))
                else:
                    targets.append([larg])
            elif isinstance(larg, Iterable):
                targets.append(larg)
            else:
                raise TypeError(
                    f"Invalid target type for {arg} of type {type(larg)}")

            for tg in zip(*targets):
                if operation == mc.Barrier:
                    self.instructions.append(
                        Instruction(mc.Barrier(N), (*tg,), (), ()))
                else:
                    self.instructions.append(
                        Instruction(
                            operation, (*tg[:N],), (*
                                                    tg[N: N + M],), (*tg[N + M:],)
                        )
                    )
        else:
            self.instructions.append(Instruction(operation))
        return self

    def _emplace_operation(self, op, regs):
        lr = len(regs)
        lq = op.num_qregs
        lc = op.num_cregs
        lz = op._num_zregs

        if lr != lq + lc + lz:
            raise ValueError(
                f"Wrong number of registers. Expected {lq} quantum + {lc} classical + {lz} z, got {lr}"
            )

        qr = op.qregsizes
        for i in range(lq):
            if len(regs[i]) != qr[i]:
                raise ValueError(
                    f"Wrong size for {i}th quantum register. Expected {qr[i]} but got {len(regs[i])}."
                )

        cr = op.cregsizes
        for i in range(lc):
            if len(regs[lq + i]) != cr[i]:
                raise ValueError(
                    f"Wrong size for {i}th classical register. Expected {cr[i]} but got {len(regs[lq + i])}."
                )

        zr = op.zregsizes
        for i in range(lz):
            if len(regs[lq + lc + i]) != zr[i]:
                raise ValueError(
                    f"Wrong size for {i}th z-register. Expected {zr[i]} but got {len(regs[lq + lc + i])}."
                )

        targets = []
        for reg in regs:
            targets.extend(reg)

        self.push(op, *targets)

        return self

    def emplace(self, op, *regs):
        """
        Constructs and adds an Operation to the end of the circuit.

        It is useful to add to the circuit operations that are dependent on the
        number of qubits.

        Arguments:
            operation (Type subclass of Operation): the type of operation to
                add.
            args (vararg of list): A variable number of arguments compriseing a
                list of parameters (if the operation is parametric), one list
                of qubits for each quantum register,  and one list of bits of
                every classical register supported.

        Examples:

            >>> from mimiqcircuits import *
            >>> c = Circuit()
            >>> c.emplace(GateX(), [0])
            1-qubit circuit with 1 instructions:
            └── X @ q[0]
            <BLANKLINE>
            >>> c.emplace(GateRX(0.2), [0])
            1-qubit circuit with 2 instructions:
            ├── X @ q[0]
            └── RX(0.2) @ q[0]
            <BLANKLINE>
            >>> c.emplace(QFT(), range(10))
            10-qubit circuit with 3 instructions:
            ├── X @ q[0]
            ├── RX(0.2) @ q[0]
            └── QFT @ q[0,1,2,3,4,5,6,7,8,9]
            <BLANKLINE>

        """

        if isinstance(op, mc.LazyExpr):
            self._emplace_operation(op(*[len(reg) for reg in regs]), regs)
        elif isinstance(op, mc.Parallel):
            if any((isinstance(reg, Iterable) and len(reg) != 1) for reg in regs):
                raise ValueError(
                    "Each iterable should contain exactly one qubit.")
            self.push(op, *regs)
        elif isinstance(op, mc.Operation):
            if isinstance(op, mc.AbstractOperator) and not isinstance(op, mc.Gate):
                raise ValueError(
                    "Cannot emplace an abstract operator that is not a Gate into the circuit."
                )
            self._emplace_operation(op, regs)
        elif isinstance(op, type) and issubclass(op, mc.Operation):
            self._emplace_operation(op, regs)
        else:
            raise TypeError("Invalid type passed to emplace")

        return self

    def insert(self, index: int, operation, *args):
        """
        Inserts an operation or another circuit at a specific index in the circuit.

        Args:
            index (int): The index at which the operation should be inserted.

            operation (Operation or Instruction): the quantum operation to add.

            args (integers or iterables): Target qubits and bits for the operation
                (not instruction), given as variable number of arguments.

        Raises:
            TypeError: If operation is not an Operation object.

            ValueError: If the number of arguments is incorrect or the
                 target qubits specified are invalid.

        Examples:
            Inserting an operation to the specify index of the circuit

            >>> from mimiqcircuits import *
            >>> c= Circuit()
            >>> c.push(GateX(), 0)
            1-qubit circuit with 1 instructions:
            └── X @ q[0]
            <BLANKLINE>
            >>> c.push(GateCX(),0,1)
            2-qubit circuit with 2 instructions:
            ├── X @ q[0]
            └── CX @ q[0], q[1]
            <BLANKLINE>
            >>> c.insert(1, GateH(), 0)
            2-qubit circuit with 3 instructions:
            ├── X @ q[0]
            ├── H @ q[0]
            └── CX @ q[0], q[1]
            <BLANKLINE>
        """
        if isinstance(operation, Circuit):
            for inst in operation.instructions:
                self.instructions.insert(index, inst)
                index += 1
        else:
            N = 0
            M = 0
            Z = 0
            L = len(args)

            if isinstance(operation, Instruction):
                if L != 0:
                    raise ValueError(
                        "No extra arguments allowed when inserting an instruction."
                    )

                self.instructions.insert(index, operation)
                return self

            if operation == mc.Barrier:
                N = L
            else:
                if not isinstance(operation, mc.Operation):
                    raise TypeError("Non Operation object passed to push.")

                # Check if the operation is an abstract operator but not a gate
                if isinstance(operation, mc.AbstractOperator) and not isinstance(
                    operation, mc.Gate
                ):
                    raise ValueError(
                        "Cannot add an abstract operator that is not a Gate to the circuit."
                    )

                N = operation.num_qubits
                M = operation.num_bits
                Z = operation.num_zvars

            if L != N + M + Z:
                raise ValueError(
                    f"Wrong number of target qubits and bits, given {L} for a {N} qubits + {M} bits operation + {Z} Z-variables."
                )

            if operation == mc.Barrier:
                self.instructions.insert(
                    index, Instruction(mc.Barrier(N), (*args,), ())
                )
            else:
                self.instructions.insert(
                    index,
                    Instruction(
                        operation, (*args[:N],), (*
                                                  args[N: N + M],), (*args[N + M:],)
                    ),
                )

        return self

    def append(self, other):
        """
        Appends all the gates of the given circuit at the end of the current
        circuit.

        Args:
            other (Circuit): the circuit to append.
        """
        instructions = None
        if isinstance(other, Circuit):
            instructions = other.instructions
        elif isinstance(other, list):
            instructions = other
        else:
            raise TypeError(
                "Only allowed to append a circuit or a list of instructions"
            )

        for inst in instructions:
            self.instructions.append(inst)

    def remove(self, index: int):
        """
        Removes an instruction at a specific index from the circuit.

        Args:
            index (int): The index of the gate to remove.

        Raises:
            IndexError: If index is out of range.
        """
        del self.instructions[index]

    def inverse(self):
        """
        Returns the inverse of the circuit.
        """
        invgates = [x.inverse() for x in self.instructions]
        invgates.reverse()
        return Circuit(invgates)

    def _decompose(self, circ):
        for instruction in self.instructions:
            instruction._decompose(circ)
        return circ

    def decompose(self):
        """
        Decompose all the gates in the circuit.

        If applied multiple times, will reduce the circuit to a basis set of U
        and CX gates.
        """
        return self._decompose(Circuit())

    def evaluate(self, d):
        c = Circuit()

        for inst in self.instructions:
            c.push(inst.evaluate(d))

        return c

    def __len__(self):
        return len(self.instructions)

    def __iter__(self):
        return iter(self.instructions)

    def __getitem__(self, index):
        if type(index) is slice:
            return Circuit(self.instructions[index])
        return self.instructions[index]

    def __str__(self, compact=False):
        lines = shutil.get_terminal_size().lines
        n = len(self)

        output = ""

        if compact and n > 0:
            nq = self.num_qubits()
            output += f"{nq}-qubit circuit with {n} instructions:\n"

            if lines - 4 <= 0:
                output += "└── ...\n"
            else:
                max_display = min(n, lines - 4)

                for g in self.instructions[: max_display - 1]:
                    output += f"├── {g}\n"

                if n > max_display:
                    output += "⋮   ⋮\n"

                g = self.instructions[-1]
                output += f"└── {g}\n"

        elif not compact and n > 0:
            nq = self.num_qubits()
            output = f"{nq}-qubit circuit with {n} instructions:"

            # iterate from the second gate
            for g in self.instructions[:-1]:
                output += f"\n├── {g}"

            g = self.instructions[-1]
            output += f"\n└── {g}"

        elif n == 0:
            output += "empty circuit"

        return output

    def __repr__(self):
        return self.__str__(compact=True)

    def __eq__(self, other):
        if not isinstance(other, Circuit):
            return False

        return self.instructions == other.instructions

    def __ne__(self, other):
        return not self.__eq__(other)

    def __class__(self):
        return Circuit

    def depth(self):
        """
        Computes the depth of the quantum circuit, including qubits, bits, and z-registers.
        """
        if self.empty() or self.num_qubits() == 0:
            return 0

        d = [0 for _ in range(self.num_qubits() +
                              self.num_bits() + self.num_zvars())]

        for g in self:
            if isinstance(g.operation, mc.Barrier):
                continue
            nq = self.num_qubits()
            nb = self.num_bits()
            optargets = (
                g.qubits
                + tuple(map(lambda x: x + nq, g.bits))
                + tuple(map(lambda x: x + nq + nb, g.zvars))
            )
            dm = max([d[t] for t in optargets])
            for t in g.qubits:
                d[t] = dm + 1
            for t in g.bits:
                d[t + nq] = dm + 1
            for t in g.zvars:
                d[t + nq + nb] = dm + 1

        return max(d)

    def copy(self):
        """Creates a shallow copy of the circuit.
            To create a full copy use deepcopy() instead.

        Returns:
            Circuit: A new Circuit object containing references to the same attributes as the original circuit
        """
        return copy.copy(self)

    def deepcopy(self):
        """Creates a copy of the object and for all its attributes

        Returns:
            Circuit: A new Circuit object fully identical the original circuit
        """
        return copy.deepcopy(self)

    def get_on_qubits(self, target_qubits):
        """
        Get instructions that involve the specified target qubits.

        Arguments:
            target_qubits (list or int): Qubits for which to retrieve instructions.

        Returns:
            Circuit: A new Circuit containing only the instructions that involve the specified qubits.
        """
        if isinstance(target_qubits, int):
            target_qubits = [target_qubits]

        selected_instructions = []

        for instruction in self.instructions:
            instruction_qubits = set(instruction.qubits + instruction.bits)
            if any(qubit in instruction_qubits for qubit in target_qubits):
                selected_instructions.append(instruction)

        return Circuit(instructions=selected_instructions)

    def saveproto(self, file):
        """
        Saves the circuit as a protobuf (binary) file.

        Arguments:
            filename (str): The name of the file to save the circuit to.

        Returns:
            int: The number of bytes written to the file.

        Examples:

            >>> from mimiqcircuits import *
            >>> from symengine import *
            >>> import tempfile
            >>> x, y = symbols("x y")
            >>> c = Circuit()
            >>> c.push(GateH(), 0)
            1-qubit circuit with 1 instructions:
            └── H @ q[0]
            <BLANKLINE>
            >>> c.push(GateXXplusYY(x**2, y),0,1)
            2-qubit circuit with 2 instructions:
            ├── H @ q[0]
            └── XXplusYY(x**2, y) @ q[0,1]
            <BLANKLINE>
            >>> c.push(Measure(),0,0)
            2-qubit circuit with 3 instructions:
            ├── H @ q[0]
            ├── XXplusYY(x**2, y) @ q[0,1]
            └── M @ q[0], c[0]
            <BLANKLINE>
            >>> tmpfile = tempfile.NamedTemporaryFile(suffix=".pb", delete=True)
            >>> c.saveproto(tmpfile.name)
            64
            >>> c.loadproto(tmpfile.name)
            2-qubit circuit with 3 instructions:
            ├── H @ q[0]
            ├── XXplusYY(x**2, y) @ q[0,1]
            └── M @ q[0], c[0]
            <BLANKLINE>

            Note:
                This example uses a temporary file to demonstrate the save and load functionality.
                You can save your file with any name at any location using:

                .. code-block:: python

                    c.saveproto("example.pb")
                    c.loadproto("example.pb")
        """
        from mimiqcircuits.proto.circuitproto import toproto_circuit

        if hasattr(file, "write"):
            return file.write(toproto_circuit(self).SerializeToString())
        else:
            try:
                with open(file, "wb") as f:
                    return f.write(toproto_circuit(self).SerializeToString())
            except TypeError:
                raise ValueError(
                    "Invalid file object. Sould be a filename of a file-like object"
                )
            except Exception as e:
                raise e

    @staticmethod
    def loadproto(file):
        """
        Loads a circuit from a protobuf (binary) file.

        Arguments:
            filename (str): The name of the file to load the circuit from.

        Returns:
            Circuit: The circuit loaded from the file.

        Note:

            Look for example in :func:`Circuit.saveproto`

        """

        from mimiqcircuits.proto import circuit_pb2
        from mimiqcircuits.proto.circuitproto import fromproto_circuit

        if isinstance(file, str):
            with open(file, "rb") as f:
                circuit_proto = circuit_pb2.Circuit()
                circuit_proto.ParseFromString(f.read())
                return fromproto_circuit(circuit_proto)
        elif hasattr(file, "read"):
            circuit_proto = circuit_pb2.Circuit()
            circuit_proto.ParseFromString(file.read())
            return fromproto_circuit(circuit_proto)
        else:
            raise ValueError(
                "Invalid file object. Sould be a filename of a file-like object"
            )

    def draw(self):
        """
        Draws the entire quantum circuit on the ASCII canvas and handles the layout of various quantum operations.

        This method iterates through all instructions in the circuit, determines the required width for each operation,
        and delegates the drawing of each operation to the appropriate specialized method based on the operation type.
        If an operation's width exceeds the available space in the current row of the canvas, the canvas is printed and
        reset to continue drawing from a new starting point.

        The method manages different operation types including control, measurement, reset, barrier, parallel, and
        conditional (if) operations using specific drawing methods from the `AsciiCircuit` class.

        Raises:
            TypeError: If any item in the circuit's instructions is not an instance of `Instruction`.
            ValueError: If an operation cannot be drawn because it exceeds the available canvas width even after a reset.

        Prints:
            The current state of the ASCII canvas, either incrementally after each operation if space runs out, or
            entirely at the end of processing all instructions.

        Returns:
            None
        """
        num_qubits = self.num_qubits()
        num_bits = self.num_bits()
        num_zvars = self.num_zvars()

        canvas = mc.AsciiCircuit()
        canvas.draw_wires(range(num_qubits), range(num_bits), range(num_zvars))

        for instruction in self.instructions:
            if not isinstance(instruction, Instruction):
                raise TypeError("Must be an Instruction")

            operation = instruction.get_operation()
            required_width = operation.asciiwidth(
                instruction.qubits,
                instruction.bits if hasattr(instruction, "bits") else [],
                instruction.zvars if hasattr(instruction, "zvars") else [],
            )

            if required_width > canvas.canvas.get_cols() - canvas.get_current_col():
                print(canvas.canvas)
                print("...")
                canvas.reset()
                canvas.draw_wires(range(num_qubits), range(
                    num_bits), range(num_zvars))

            if required_width > canvas.canvas.get_cols() - canvas.get_current_col():
                raise ValueError(
                    "Cannot draw instruction. Insufficient space on screen."
                )

            # Handle drawing based on operation type
            if isinstance(operation, mc.Control):
                canvas.draw_control(
                    operation,
                    instruction.qubits,
                    instruction.bits if hasattr(instruction, "bits") else [],
                )

            elif isinstance(operation, mc.PauliString):
                canvas.draw_paulistring(
                    operation,
                    instruction.qubits,
                    instruction.bits if hasattr(instruction, "bits") else [],
                    instruction.zvars if hasattr(instruction, "zvars") else [],
                )

            elif isinstance(operation, mc.Reset):
                canvas.draw_reset(
                    operation,
                    instruction.qubits,
                    instruction.bits if hasattr(instruction, "bits") else [],
                    instruction.zvars if hasattr(instruction, "zvars") else [],
                )

            elif isinstance(operation, mc.Barrier):
                canvas.draw_barrier(
                    operation,
                    instruction.qubits,
                    instruction.bits if hasattr(instruction, "bits") else [],
                    instruction.zvars if hasattr(instruction, "zvars") else [],
                )

            elif isinstance(operation, mc.Parallel):
                canvas.draw_parallel(
                    operation,
                    instruction.qubits,
                    instruction.bits if hasattr(instruction, "bits") else [],
                    instruction.zvars if hasattr(instruction, "zvars") else [],
                )

            elif isinstance(operation, mc.IfStatement):
                canvas.draw_ifstatement(
                    operation,
                    instruction.qubits,
                    instruction.bits if hasattr(instruction, "bits") else [],
                    instruction.zvars if hasattr(instruction, "zvars") else [],
                )

            elif isinstance(operation, mc.Operation):
                canvas.draw_operation(
                    operation,
                    instruction.qubits,
                    instruction.bits if hasattr(instruction, "bits") else [],
                    instruction.zvars if hasattr(instruction, "zvars") else [],
                )
            else:
                # Default drawing method for general operations
                canvas.draw_instruction(instruction)

        print(canvas.canvas)
        return None

    def specify_operations(self):
        """
        Summarizes the types and numbers of operations in the circuit.

        This function inspects each instruction in the circuit and categorizes it by
        the number of qubits, bits, and z-variables involved in the operation. It
        then prints a summary of the total number of operations in the circuit and
        a breakdown of the number of operations grouped by their type.

        Examples:

            >>> from mimiqcircuits import *
            >>> c = Circuit()

            Add a Pauli-X (GateX) gate on qubit 0

            >>> c.push(GateX(), 0)
            1-qubit circuit with 1 instructions:
            └── X @ q[0]
            <BLANKLINE>

            Add a Controlled-NOT (CX) gate with control qubit 0 and target qubit 1

            >>> c.push(GateCX(), 0, 1)
            2-qubit circuit with 2 instructions:
            ├── X @ q[0]
            └── CX @ q[0], q[1]
            <BLANKLINE>

            Add a Measurement operation on qubit 0, storing the result in bit 0

            >>> c.push(Measure(), 0, 0)
            2-qubit circuit with 3 instructions:
            ├── X @ q[0]
            ├── CX @ q[0], q[1]
            └── M @ q[0], c[0]
            <BLANKLINE>

            Add an ExpectationValue operation with GateX on qubit 1, storing the result in z-variable 2.

            >>> c.push(ExpectationValue(GateX()), 1, 2)
            2-qubit circuit with 4 instructions:
            ├── X @ q[0]
            ├── CX @ q[0], q[1]
            ├── M @ q[0], c[0]
            └── ⟨X⟩ @ q[1], z[2]
            <BLANKLINE>

            Print a summary of the types and numbers of operations

            >>> c.specify_operations()
            Total number of operations: 4
            ├── 1 x 1_qubits
            ├── 1 x 2_qubits
            ├── 1 x 1_qubits & 1_bits
            └── 1 x 1_qubits & 1_zvars
        """
        counts = {}
        for i in self.instructions:
            nq = len(i._qubits)
            nb = len(i._bits)
            nz = len(i._zvars)

            key = ""

            if nq > 0:
                key += f"{nq}_qubits"

            if nb > 0:
                if key:
                    key += " & "
                key += f"{nb}_bits"
            if nz > 0:
                if key:
                    key += " & "
                key += f"{nz}_zvars"

            counts[key] = counts.get(key, 0) + 1

        total_operations = sum(counts.values())
        print(f"Total number of operations: {total_operations}")

        count_items = list(counts.items())
        for idx, (key, count) in enumerate(count_items):
            if idx == len(count_items) - 1:
                print(f"└── {count} x {key}")
            else:
                print(f"├── {count} x {key}")

    def is_symbolic(self):
        """
        Check whether the circuit contains any symbolic (unevaluated) parameters.

        This method examines each instruction in the circuit to determine if any parameter remains
        symbolic (i.e., unevaluated). It recursively checks through each instruction and its nested
        operations, if any.

        Returns:
            bool: True if any parameter is symbolic (unevaluated), False if all parameters are fully evaluated.

        Examples:

            >>> from mimiqcircuits import *
            >>> from symengine import *
            >>> x, y = symbols("x y")
            >>> c = Circuit()
            >>> c.push(GateH(), 0)
            1-qubit circuit with 1 instructions:
            └── H @ q[0]
            <BLANKLINE>
            >>> c.is_symbolic()
            False
            >>> c.push(GateP(x), 0)
            1-qubit circuit with 2 instructions:
            ├── H @ q[0]
            └── P(x) @ q[0]
            <BLANKLINE>
            >>> c.is_symbolic()
            True
            >>> c = c.evaluate({x: 1, y: 2})
            >>> c
            1-qubit circuit with 2 instructions:
            ├── H @ q[0]
            └── P(1) @ q[0]
            <BLANKLINE>
            >>> c.is_symbolic()
            False
        """
        return any(
            instruction.operation.is_symbolic() for instruction in self.instructions
        )

    def add_noise_to_gate_single(
        self, g, kraus: Union[mc.krauschannel, mc.Gate], before=False
    ):
        """
        Adds a noise operation `kraus` before or after every instance of a given operation `g`.

        The noise operation `kraus` can be a Kraus channel or a gate and will act on the same qubits
        as the operation `g` to which it is being added.

        Args:
            g (Operation): The operation to which the noise will be added.
            kraus (Operation): The noise operation to be added, which can be a Kraus channel or any valid gate.
            before (bool, optional): If True, the noise is added before the operation `g`. Default is False.

        Raises:
            ValueError: If the noise operation is the same as the operation `g`, to avoid recursion.

        See also:
            :func:`add_noise`: Adds noise to multiple operations at once or in a parallel block.
        """
        if not isinstance(g, mc.Operation):
            raise ValueError(f"{g} must be an Operation")

        if not isinstance(kraus, (mc.krauschannel, mc.Gate)):
            raise ValueError(
                f"{kraus} is not of type {mc.krauschannel} or {mc.Gate}")

        if g == kraus:
            raise ValueError(
                "Noise can't be the same as gate, otherwise recursion problem."
            )

        rel = 0 if before else 1

        i = 0
        while i < len(self.instructions):
            if self.instructions[i].operation == g:
                qubits = self.instructions[i].qubits
                self.insert(i + rel, kraus, *qubits)
                i += 1
            i += 1

        return self

    def add_noise_to_gate_parallel(
        self, g, kraus: Union[mc.krauschannel, mc.Gate], before=False
    ):
        """
        Adds a block of noise operations `kraus` after/before every block of a given operation `g`.

        The function identifies blocks of consecutive transversal operations of type `g` and
        adds a block of transversal noise operations `kraus` after each such block. The noise operation
        `kraus` can be a Kraus channel or a gate and will act on the same qubits as the operation `g` to which it is being added.

        Args:
            g (Operation): The operation to which the noise will be added.
            kraus (Operation): The noise operation to be added, which can be a Kraus channel or any valid gate.
            before (bool, optional): If True, the noise is added before the operation `g`. Default is False.

        Raises:
            ValueError: If the noise operation is the same as the operation `g`, to avoid recursion.

        See also:
            :func:`add_noise`: Adds noise to multiple operations at once or in a parallel block.
        """

        if not isinstance(g, mc.Operation):
            raise ValueError(f"{g} must be an Operation")

        if not isinstance(kraus, (mc.krauschannel, mc.Gate)):
            raise ValueError(
                f"{kraus} is not of type {mc.krauschannel} or {mc.Gate}")

        if g == kraus:
            raise ValueError(
                "Noise can't be the same as gate, otherwise recursion problem."
            )

        i = 0
        while i < len(self.instructions):
            inds = [i]
            qubits = set(self.instructions[i].qubits)

            if self.instructions[i].operation == g:
                j = i + 1
                while (
                    j < len(self.instructions)
                    and self.instructions[j].operation == g
                    and not qubits.intersection(set(self.instructions[j].qubits))
                ):
                    inds.append(j)
                    qubits.update(self.instructions[j].qubits)
                    j += 1

                for rel, j in enumerate(inds):
                    if before:
                        self.insert(i + rel, kraus, *
                                    self.instructions[j].qubits)
                    else:
                        self.insert(
                            inds[-1] + 1 + rel, kraus, *
                            self.instructions[j].qubits
                        )

                i += len(inds)  # Noise block shifting
            i += len(inds)  # Gate shifting

        return self

    def add_noise(
        self,
        g: Union[mc.Operation, List[mc.Operation]],
        kraus: Union[mc.krauschannel, mc.Gate, List[mc.Gate], List[mc.krauschannel]],
        before: Union[bool, List[bool]] = False,
        parallel: Union[bool, List[bool]] = False,
    ):
        """
        Adds a noise operation `kraus` to every instance of the operation `g` in the circuit.

        The noise operation `kraus` can be a Kraus channel or a gate and will act on the same qubits
        as the operation `g` to which it is being added.

        The operations `g` and `kraus` must act on the same number of qubits.

        Args:
            g (Operation or list of Operation): The operation(s) to which the noise will be added.
            kraus (krauschannel or list of krauschannel): The noise operation(s) to be added.
            before (bool or list of bool, optional): If True, the noise is added before the operation. Default is False.
            parallel (bool or list of bool, optional): If True, noise is added as a block. Default is False.

        Raises:
            ValueError: If `g` and `kraus` are not of the same length, or if their number of qubits differ.
            TypeError: If `before` or `parallel` are not a bool or a list of bool.

        Returns:
            Circuit: The modified circuit with the noise added.

        Examples:

        Adding noise sequentially (not parallel):

        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(GateH(), [1,2,3])
        4-qubit circuit with 3 instructions:
        ├── H @ q[1]
        ├── H @ q[2]
        └── H @ q[3]
        <BLANKLINE>
        >>> c.add_noise(GateH(), AmplitudeDamping(0.2))
        4-qubit circuit with 6 instructions:
        ├── H @ q[1]
        ├── AmplitudeDamping(0.2) @ q[1]
        ├── H @ q[2]
        ├── AmplitudeDamping(0.2) @ q[2]
        ├── H @ q[3]
        └── AmplitudeDamping(0.2) @ q[3]
        <BLANKLINE>

        Adding noise in parallel:

        >>> c = Circuit()
        >>> c.push(GateH(), [1, 2, 3])
        4-qubit circuit with 3 instructions:
        ├── H @ q[1]
        ├── H @ q[2]
        └── H @ q[3]
        <BLANKLINE>
        >>> c.add_noise(GateH(), AmplitudeDamping(0.2), parallel=True)
        4-qubit circuit with 6 instructions:
        ├── H @ q[1]
        ├── H @ q[2]
        ├── H @ q[3]
        ├── AmplitudeDamping(0.2) @ q[1]
        ├── AmplitudeDamping(0.2) @ q[2]
        └── AmplitudeDamping(0.2) @ q[3]
        <BLANKLINE>

        Parallel will not work if gates aren't transversal.

        >>> c = Circuit()
        >>> c.push(GateCZ(), 1, range(2,5))
        5-qubit circuit with 3 instructions:
        ├── CZ @ q[1], q[2]
        ├── CZ @ q[1], q[3]
        └── CZ @ q[1], q[4]
        <BLANKLINE>
        >>> c.add_noise(GateCZ(), Depolarizing2(0.1), parallel=True)
        5-qubit circuit with 6 instructions:
        ├── CZ @ q[1], q[2]
        ├── Depolarizing(0.1) @ q[1,2]
        ├── CZ @ q[1], q[3]
        ├── Depolarizing(0.1) @ q[1,3]
        ├── CZ @ q[1], q[4]
        └── Depolarizing(0.1) @ q[1,4]
        <BLANKLINE>

        Adding noise before measurement (The `before=True` option is mostly used for `Measure`):

        >>> c = Circuit()
        >>> c.push(Measure(), [1, 2, 3], [1, 2, 3])
        4-qubit circuit with 3 instructions:
        ├── M @ q[1], c[1]
        ├── M @ q[2], c[2]
        └── M @ q[3], c[3]
        <BLANKLINE>
        >>> c.add_noise(Measure(), PauliX(0.1), before=True)
        4-qubit circuit with 6 instructions:
        ├── PauliX(0.1) @ q[1]
        ├── M @ q[1], c[1]
        ├── PauliX(0.1) @ q[2]
        ├── M @ q[2], c[2]
        ├── PauliX(0.1) @ q[3]
        └── M @ q[3], c[3]
        <BLANKLINE>

        Adding unitary gates as noise in the same way:

        >>> c = Circuit()
        >>> c.push(GateH(), [1, 2, 3])
        4-qubit circuit with 3 instructions:
        ├── H @ q[1]
        ├── H @ q[2]
        └── H @ q[3]
        <BLANKLINE>
        >>> c.add_noise(GateH(), GateRX(0.01))
        4-qubit circuit with 6 instructions:
        ├── H @ q[1]
        ├── RX(0.01) @ q[1]
        ├── H @ q[2]
        ├── RX(0.01) @ q[2]
        ├── H @ q[3]
        └── RX(0.01) @ q[3]
        <BLANKLINE>

        """

        # Ensure g and kraus are treated as lists
        g = g if isinstance(g, list) else [g]
        kraus = kraus if isinstance(kraus, list) else [kraus]

        # Check for length mismatch between g and kraus
        if len(g) != len(kraus):
            raise ValueError(
                "Vectors of operations and noise channels must have the same length."
            )

        nops = len(g)

        # Check types of 'before' and 'parallel' and ensure they are the correct length
        if not isinstance(before, (bool, list)):
            raise TypeError(
                "Parameter 'before' has to be a bool or a list of bool.")
        if isinstance(before, list) and len(before) != nops:
            raise ValueError(
                "Vector 'before' must have the same length as the vector of operations."
            )

        if not isinstance(parallel, (bool, list)):
            raise TypeError(
                "Parameter 'parallel' has to be a bool or a list of bool.")
        if isinstance(parallel, list) and len(parallel) != nops:
            raise ValueError(
                "Vector 'parallel' must have the same length as the vector of operations."
            )

        # Vectorize optional parameters if they are not already lists
        if isinstance(before, bool):
            before = [before] * nops
        if isinstance(parallel, bool):
            parallel = [parallel] * nops

        # Validate each operation and noise channel pair
        for operation, noise in zip(g, kraus):
            if not isinstance(operation, mc.Operation):
                raise ValueError(
                    f"{operation} must be an instance of mc.Operation")
            # Check for individual noise type in the list case
            if isinstance(noise, list):
                for n in noise:
                    if not isinstance(n, (mc.krauschannel, mc.Gate)):
                        raise ValueError(
                            f"{n} is not of type {mc.krauschannel} or {mc.Gate}"
                        )
            elif not isinstance(noise, (mc.krauschannel, mc.Gate)):
                raise ValueError(
                    f"{noise} is not of type {mc.krauschannel} or {mc.Gate}"
                )

            if operation.num_qubits != noise.num_qubits:
                raise ValueError(
                    "Noise channel and operation must have the same number of target qubits"
                )

        # Apply noise for each pair
        for operation, noise, add_before, add_parallel in zip(
            g, kraus, before, parallel
        ):
            if add_parallel:
                self.add_noise_to_gate_parallel(
                    operation, noise, before=add_before)
            else:
                self.add_noise_to_gate_single(
                    operation, noise, before=add_before)

        return self

    def sample_mixedunitaries(self, rng=None, ids=False):
        """
            sample_mixedunitaries(rng=None, ids=False)

        Samples one unitary gate for each mixed unitary Kraus channel in the circuit.

        This is possible because for mixed unitary noise channels, the probabilities of each
        Kraus operator are fixed (state-independent).

        Note: This function is internally called (before applying any gate) when executing
        a circuit with noise using trajectories. It can also be used to generate samples
        of circuits without running them.

        See also:
            - :func:`krauschannel.ismixedunitary`
            - :class:`MixedUnitary`

        Args:
            rng (optional): Random number generator. If not provided, Python's default
                random number generator is used.
            ids (optional): Boolean, default=False. Determines whether to include identity
                Kraus operators in the sampled circuit. If True, identity gates are added
                to the circuit; otherwise, they are omitted. Usually, most selected Kraus
                operators will be identity gates.

        Returns:
            Circuit: A copy of the circuit with every mixed unitary Kraus channel replaced
            by one of the unitary gates of the channel. Identity gates are omitted unless
            `ids=True`.

        Examples:
            Gates and non-mixed-unitary Kraus channels remain unchanged:

            >>> from mimiqcircuits import *
            >>> c = Circuit()
            >>> c.push(GateH(), [1, 2, 3])
            4-qubit circuit with 3 instructions:
            ├── H @ q[1]
            ├── H @ q[2]
            └── H @ q[3]
            <BLANKLINE>
            >>> c.push(Depolarizing1(0.5), [1, 2, 3])
            4-qubit circuit with 6 instructions:
            ├── H @ q[1]
            ├── H @ q[2]
            ├── H @ q[3]
            ├── Depolarizing(0.5) @ q[1]
            ├── Depolarizing(0.5) @ q[2]
            └── Depolarizing(0.5) @ q[3]
            <BLANKLINE>
            >>> c.push(AmplitudeDamping(0.5), [1, 2, 3])
            4-qubit circuit with 9 instructions:
            ├── H @ q[1]
            ├── H @ q[2]
            ├── H @ q[3]
            ├── Depolarizing(0.5) @ q[1]
            ├── Depolarizing(0.5) @ q[2]
            ├── Depolarizing(0.5) @ q[3]
            ├── AmplitudeDamping(0.5) @ q[1]
            ├── AmplitudeDamping(0.5) @ q[2]
            └── AmplitudeDamping(0.5) @ q[3]
            <BLANKLINE>

            >>> rng = random.Random(42)

            >>> new_circuit = c.sample_mixedunitaries(rng=rng, ids=True)
            >>> print(new_circuit)
            4-qubit circuit with 9 instructions:
            ├── H @ q[1]
            ├── H @ q[2]
            ├── H @ q[3]
            ├── X @ q[1]
            ├── I @ q[2]
            ├── I @ q[3]
            ├── AmplitudeDamping(0.5) @ q[1]
            ├── AmplitudeDamping(0.5) @ q[2]
            └── AmplitudeDamping(0.5) @ q[3]

            By default, identities are not included:

            >>> new_circuit = c.sample_mixedunitaries(rng=rng)
            >>> print(new_circuit)
            4-qubit circuit with 8 instructions:
            ├── H @ q[1]
            ├── H @ q[2]
            ├── H @ q[3]
            ├── Y @ q[2]
            ├── Y @ q[3]
            ├── AmplitudeDamping(0.5) @ q[1]
            ├── AmplitudeDamping(0.5) @ q[2]
            └── AmplitudeDamping(0.5) @ q[3]

            Different calls to the function generate different results:

            >>> new_circuit = c.sample_mixedunitaries(rng=rng)
            >>> print(new_circuit)
            4-qubit circuit with 7 instructions:
            ├── H @ q[1]
            ├── H @ q[2]
            ├── H @ q[3]
            ├── Z @ q[1]
            ├── AmplitudeDamping(0.5) @ q[1]
            ├── AmplitudeDamping(0.5) @ q[2]
            └── AmplitudeDamping(0.5) @ q[3]

            >>> new_circuit = c.sample_mixedunitaries(rng=rng)
            >>> print(new_circuit)
            4-qubit circuit with 7 instructions:
            ├── H @ q[1]
            ├── H @ q[2]
            ├── H @ q[3]
            ├── X @ q[3]
            ├── AmplitudeDamping(0.5) @ q[1]
            ├── AmplitudeDamping(0.5) @ q[2]
            └── AmplitudeDamping(0.5) @ q[3]
        """

        if rng is None:
            rng = random()

        scirc = Circuit()

        for inst in self.instructions:
            op = inst.get_operation()

            if isinstance(op, mc.krauschannel) and op.ismixedunitary():
                cumulative_probs = op.unwrappedcumprobabilities()

                r = rng.random()

                # Use `next` to find the index of the first cumulative probability greater than `r`
                index = next(i for i, p in enumerate(
                    cumulative_probs) if p > r)

                gate = op.unitarygates()[index]

                if ids or not gate.isidentity():
                    scirc.push(gate, *inst.qubits)
            else:
                scirc.push(inst)

        return scirc


# export the cirucit classes
__all__ = ["Instruction", "Circuit"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.gates.standard.sx import GateSX, GateSXDG
from mimiqcircuits.operations.gates.standard.hadamard import GateH
from mimiqcircuits.operations.gates.standard.deprecated import GateU1
import mimiqcircuits.operations.control as mctrl
from symengine import pi


class GateCSX(mctrl.Control):
    r"""Two qubit Controled-SX gate.

    By convention, the first qubit is the control and second one is the
    targets.


    **Matrix representation:**

    .. math::

        \operatorname{CSX} =\begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & \frac{1+i}{2} & \frac{1-i}{2} \\
            0 & 0 & \frac{1-i}{2} & \frac{1+i}{2}
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateCSX(), GateCSX().num_controls, GateCSX().num_targets, GateCSX().num_qubits
        (CSX, 1, 1, 2)
        >>> GateCSX().matrix()
        [1.0, 0, 0, 0]
        [0, 1.0, 0, 0]
        [0, 0, 0.5 + 0.5*I, 0.5 - 0.5*I]
        [0, 0, 0.5 - 0.5*I, 0.5 + 0.5*I]
        <BLANKLINE>
        >>> c = Circuit().push(GateCSX(), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── CSX @ q[0], q[1]
        <BLANKLINE>
        >>> GateCSX().power(2), GateCSX().inverse()
        (CX, CSX†)
        >>> GateCSX().decompose()
        2-qubit circuit with 3 instructions:
        ├── H @ q[1]
        ├── CU1((1/2)*pi) @ q[0], q[1]
        └── H @ q[1]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(1, GateSX())

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = qubits

        circ.push(GateH(), b)
        circ.push(GateU1(pi / 2).control(1), a, b)
        circ.push(GateH(), b)

        return circ


class GateCSXDG(mctrl.Control):
    r"""Two qubit :math:`{CSX}^\dagger` gate.

    **Matrix representation:**

    .. math::
        \operatorname{CSX}^{\dagger} =\begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & \frac{1-i}{2} & \frac{1+i}{2} \\
            0 & 0 & \frac{1+i}{2} & \frac{1-i}{2}
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateCSXDG(), GateCSXDG().num_controls, GateCSXDG().num_targets, GateCSXDG().num_qubits
        (CSX†, 1, 1, 2)
        >>> GateCSXDG().matrix()
        [1.0, 0, 0, 0]
        [0, 1.0, 0, 0]
        [0, 0, 0.5 - 0.5*I, 0.5 + 0.5*I]
        [0, 0, 0.5 + 0.5*I, 0.5 - 0.5*I]
        <BLANKLINE>
        >>> c = Circuit().push(GateCSXDG(), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── CSX† @ q[0], q[1]
        <BLANKLINE>
        >>> GateCSXDG().power(2), GateCSXDG().inverse()
        (C(SX†**2), CSX)
        >>> GateCSXDG().decompose()
        2-qubit circuit with 3 instructions:
        ├── H @ q[1]
        ├── CU1((-1/2)*pi) @ q[0], q[1]
        └── H @ q[1]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(1, GateSXDG())

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = range(self.num_qubits)

        circ.push(GateH(), b)
        circ.push(GateU1(-pi / 2).control(1), a, b)
        circ.push(GateH(), b)

        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import symengine as se
from mimiqcircuits.operations.krauschannel import krauschannel
import mimiqcircuits as mc


def checknormalizlength(normaliz, length):
    if len(normaliz) != length:
        raise ValueError(f"normaliz must be a list or tuple of length {length}.")


class ProjectiveNoiseX(krauschannel):
    r"""
    ProjectiveNoiseX()

    Single qubit projection noise onto an X Pauli basis.

    This channel is defined by the Kraus operators:

    .. math::
        E_1 = |-\rangle \langle-|, \quad E_2 = |+\rangle \langle+|,

    where :math:`\ket{+}` and :math:`\ket{-}` are the eigenstates of Pauli `X`.

    See also:
        - :class:`ProjectiveNoise`
        - :class:`ProjectiveNoiseY`
        - :class:`ProjectiveNoiseZ`
    """

    _name = "ProjectiveNoiseX"
    _num_qubits = 1
    _num_bits = 0
    _num_zvars = 0
    _qregsizes = [1]

    def opname(self):
        return "ProjectiveNoiseX"

    def krausoperators(self):
        return [mc.ProjectorX0(), mc.ProjectorX1()]

    def __str__(self):
        return f"{self._name}"


class ProjectiveNoiseY(krauschannel):
    r"""
    ProjectiveNoiseY()

    Single qubit projection noise onto a Y Pauli basis.

    This channel is defined by the Kraus operators:

    .. math::
        E_1 = |Y0\rangle \langle Y0|, \quad E_2 = |Y1\rangle \langle Y1|,

    where :math:`\ket{Y0}` and :math:`\ket{Y1}` are the eigenstates of Pauli `Y`.

    See also:
        - :class:`ProjectiveNoise`
        - :class:`ProjectiveNoiseX`
        - :class:`ProjectiveNoiseZ`
    """

    _name = "ProjectiveNoiseY"
    _num_qubits = 1
    _num_bits = 0
    _num_zvars = 0
    _qregsizes = [1]

    def opname(self):
        return "ProjectiveNoiseY"

    def krausoperators(self):
        return [mc.ProjectorY0(), mc.ProjectorY1()]

    def __str__(self):
        return f"{self._name}"


class ProjectiveNoiseZ(krauschannel):
    r"""
    ProjectiveNoiseZ()

    Single qubit projection noise onto a Z Pauli basis.

    This channel is defined by the Kraus operators:

    .. math::
        E_1 = |0\rangle \langle Z0|, \quad E_2 = |1\rangle \langle Z1|,

    where :math:`\ket{0}` and :math:`\ket{1}` are the eigenstates of Pauli `Z`.

    See also:
        - :class:`ProjectiveNoise`
        - :class:`ProjectiveNoiseX`
        - :class:`ProjectiveNoiseY`
    """

    _name = "ProjectiveNoiseZ"
    _num_qubits = 1
    _num_bits = 0
    _num_zvars = 0
    _qregsizes = [1]

    def opname(self):
        return "ProjectiveNoiseZ"

    def __str__(self):
        return f"{self._name}"

    def krausoperators(self):
        return [mc.Projector0(), mc.Projector1()]



class ProjectiveNoise(krauschannel):
    r"""
    ProjectiveNoise(basis)

    Single qubit projection noise onto a Pauli basis.

    This channel is defined by the Kraus operators:

    .. math::
        E_1 = |\alpha\rangle \langle\alpha|, \quad E_2 = |\beta\rangle \langle\beta|,

    where the states :math:`|\alpha\rangle` and :math:`|\beta\rangle` are the +1 and -1
    eigenstates of a Pauli operator. Specifically, they correspond to
    :math:`\{ |0\rangle, |1\rangle \}` (Z basis),
    :math:`\{ |+\rangle, |-\rangle \}` (X basis),
    or :math:`\{ |y+\rangle, |y-\rangle \}` (Y basis).

    This operation is similar to measuring in the corresponding basis (X, Y, or Z),
    except that the outcome of the measurement is not stored, i.e., there is a loss of information.

    Parameters:
        basis (str): String or character that selects the Pauli basis, `"X"`, `"Y"`, or `"Z"`.

    Examples:
        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(ProjectiveNoise("Z"), 1)
        2-qubit circuit with 1 instructions:
        └── ProjectiveNoiseZ @ q[1]
        <BLANKLINE>

    The Kraus matrices are given by:

        >>> ProjectiveNoise("X").krausmatrices()
        [[0.5, 0.5]
        [0.5, 0.5]
        , [0.5, -0.5]
        [-0.5, 0.5]
        ]

        >>> ProjectiveNoise("Y").krausmatrices()
        [[0.5, -0.0 - 0.5*I]
        [0.0 + 0.5*I, 0.5]
        , [0.5, 0.0 + 0.5*I]
        [-0.0 - 0.5*I, 0.5]
        ]

        >>> ProjectiveNoise("Z").krausmatrices()
        [[1.0, 0]
        [0, 0]
        , [0, 0]
        [0, 1.0]
        ]
    """

    _name = "ProjectiveNoise"
    _num_qubits = 1
    _qregsizes = [1]
    _cregsizes = ()

    def __new__(self, basis="Z"):
        if basis == "X":
            return ProjectiveNoiseX()
        elif basis == "Y":
            return ProjectiveNoiseY()
        elif basis == "Z":
            return ProjectiveNoiseZ()
        else:
            raise ValueError(
                "Invalid basis for Projective noise. Must be 'X', 'Y', or 'Z'."
            )

    def krausmatrices(self):
        return [se.Matrix(kraus.matrix().tolist()) for kraus in self.krausoperators()]

    def __str__(self):
        return f"{self._name}"


__all__ = [
    "ProjectiveNoise",
    "ProjectiveNoiseX",
    "ProjectiveNoiseY",
    "ProjectiveNoiseZ",
]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from mimiqcircuits.proto import qcsresults_pb2
from mimiqcircuits.proto import bitvector_pb2

import mimiqcircuits.qcsresults as mq

from mimiqcircuits.bitstrings import BitString
from bitarray import frozenbitarray


def bitvec_to_bytes(bv):
    b = bytearray()
    for i in range(0, len(bv), 8):
        val = 0
        for j in range(8):
            if i + j >= len(bv):
                break
            val += bv[i + j] << j
        b.append(val)
    return b


def bytes_to_bitvec(b, n=None):
    bitstring = ""
    for val in b:
        for i in range(8):
            bitstring += str((val >> i) & 1)
    if n is not None:
        return frozenbitarray(bitstring[:n])
    else:
        return frozenbitarray(bitstring)


def toproto_qcsr(s):
    from mimiqcircuits.proto.bitvector import toproto_bitvector

    qcs_results = qcsresults_pb2.QCSResults()
    qcs_results.simulator = s.simulator
    qcs_results.version = s.version
    qcs_results.fidelities.extend(s.fidelities)
    qcs_results.avggateerrors.extend(s.avggateerrors)

    for cstate in s.cstates:
        qcs_results.cstates.extend([toproto_bitvector(cstate)])

    for zstate in s.zstates:
        qcs_results.zstates.extend([toproto_complexvector(zstate)])

    for key, value in s.timings.items():
        qcs_results.timings[key] = value

    amplitude_entries = [
        toproto_amplitude(key, value) for key, value in s.amplitudes.items()
    ]
    qcs_results.amplitudes.extend(amplitude_entries)

    return qcs_results


def fromproto_qcsr(s):
    from mimiqcircuits.proto.bitvector import fromproto_bitvector

    qcs_results = mq.QCSResults(
        s.simulator,
        s.version,
        s.fidelities,
        s.avggateerrors,
        [fromproto_bitvector(cstate) for cstate in s.cstates],
        [fromproto_complexvector(zstate) for zstate in s.zstates],
        amplitudes={
            key: value
            for amplitude_entry in s.amplitudes
            for key, value in [fromproto_amplitude(amplitude_entry)]
        },
        timings=dict(s.timings),
    )

    return qcs_results


def toproto_amplitude(key, value):
    key_bit_array = frozenbitarray(key)
    entry = qcsresults_pb2.AmplitudeEntry(
        key=bitvector_pb2.BitVector(len=len(key_bit_array), data=bytes(key_bit_array)),
        val=toproto_complexdouble(value),
    )
    return entry


def fromproto_amplitude(amplitude_entry):
    from mimiqcircuits.proto.bitvector import fromproto_bitvector

    key = BitString(fromproto_bitvector(amplitude_entry.key))
    value = fromproto_complexdouble(amplitude_entry.val)
    return key, value


def toproto_complexvector(v):
    q = [toproto_complexdouble(x) for x in v]
    return qcsresults_pb2.ComplexVector(data=q)


def fromproto_complexvector(v):
    q = [fromproto_complexdouble(x) for x in v.data]
    return q


def toproto_complexdouble(v):
    return qcsresults_pb2.ComplexDouble(real=v.real, imag=v.imag)


def fromproto_complexdouble(v):
    return complex(v.real, v.imag)


__all__ = ["toproto_qcsr", "fromproto_qcsr"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc
import mimiqcircuits.lazy as lz
from mimiqcircuits.printutils import print_wrapped_parens
from mimiqcircuits.operations.gates.gate import Gate


class Inverse(Gate):
    """Inverse of the wrapped quantum operation.

    The inversion is not performed right away, but only when the circuit is
    cached or executed.

    .. warning::

        Users should not use `Inverse` directly but rather the `inverse` method,
        which performs all the necessary simplifications (e.g., `op.inverse().inverse() == op`)

    Examples:
        >>> from mimiqcircuits import *
        >>> Inverse(GateP(1)).matrix()
        [1.0, 0]
        [0, 0.54030230586814 - 0.841470984807897*I]
        <BLANKLINE>
        >>> c = Circuit()
        >>> c.push(Inverse(GateP(1)), 1)
        2-qubit circuit with 1 instructions:
        └── P(1)† @ q[1]
        <BLANKLINE>
    """

    _name = "Inverse"
    _num_qubits = None

    _num_bits = 0
    _num_cregs = 0

    _op = None

    def __init__(self, operation, *args, **kwargs):
        if isinstance(operation, type) and issubclass(operation, mc.Gate):
            op = operation(*args, **kwargs)
        elif isinstance(operation, mc.Gate):
            op = operation
        else:
            raise ValueError("Operation must be an Gate object or type.")


        if op.num_bits != 0:
            raise ValueError("Cannot inverte operation with classical bits.")

        self.op = op
        super().__init__()
        self._num_qubits = op.num_qubits
        self._num_qregs = op._num_qregs
        self._qregsizes = op._qregsizes
        self._parnames = op.parnames

    def __str__(self):
        return f"{print_wrapped_parens(self.op)}†"

    def iswrapper(self):
        return True

    def inverse(self):
        return self.op

    def getparams(self):
        return self.op.getparams()

    def _power(self, pwr):
        return mc.Power(self, pwr)

    def power(self, *args):
        if len(args) == 0:
            return lz.power(self)
        elif len(args) == 1:
            pwr = args[0]
            return self._power(pwr)
        else:
            raise ValueError("Invalid number of arguments.")

    def __pow__(self, pwr):
        return self.power(pwr)

    def control(self, *args):
        if len(args) == 0:
            return lz.control(self)
        elif len(args) == 1:
            num_controls = args[0]
            return mc.Control(num_controls, self)
        else:
            raise ValueError("Invalid number of arguments.")

    def parallel(self, *args):
        if len(args) == 0:
            return lz.parallel(self)
        elif len(args) == 1:
            num_repeats = args[0]
            return mc.Parallel(num_repeats, self)
        else:
            raise ValueError("Invalid number of arguments.")
    
    def _matrix(self):
        return self.op.matrix().inv()

    def evaluate(self, d):
        return self.op.evaluate(d).inverse()

    def _decompose(self, circ, qubits, bits, zvars):
        newc = self.op._decompose(mc.Circuit(), qubits, bits, zvars).inverse()
        circ.append(newc)
        return circ


__all__ = ["Inverse"]



#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits import AbstractOperator
import symengine as se
import mimiqcircuits as mc


class SigmaMinus(AbstractOperator):
    r"""One-qubit operator corresponding to :math:`|0 \rangle\langle 1|`.

    **Matrix Representation**

    .. math::
        \begin{pmatrix}
            0 & a\\
            0 & 0
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`SigmaPlus`

    Parameters:
        a (complex, optional): Scaling factor for the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> SigmaMinus()
        SigmaMinus(1)

        >>> SigmaMinus(0.5)
        SigmaMinus(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(SigmaMinus()), 1, 1)
        2-qubit circuit with 1 instructions:
        └── ⟨SigmaMinus(1)⟩ @ q[1], z[1]
        <BLANKLINE>
    """

    _name = "SigmaMinus"
    _num_qubits = 1
    _parnames = ()
    _qregsizes = [1]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    @property
    def parnames(self):
        return self._parnames

    @property
    def num_qubits(self):
        return self._num_qubits

    def _matrix(self):
        return se.Matrix([[0, self.a], [0, 0]])

    def opsquared(self):
        return mc.Projector1(abs(self.a) ** 2)

    def rescale(self, scale):
        return SigmaMinus(self.a * scale)

    def rescale_inplace(self, scale):
        self.a *= scale  # Modify the original object in-place
        return self

    def __str__(self):
        return f"{self.opname()}({self.a})"


class SigmaPlus(AbstractOperator):
    r"""One-qubit operator corresponding to :math:`|1 \rangle\langle 0|`.

    **Matrix Representation**

    .. math::
        \begin{pmatrix}
            0 & 0\\
            a & 0
        \end{pmatrix}

    This matrix is parametrized by `a` to allow for phases/rescaling.

    The parameter `a` is optional and is set to 1 by default.

    See Also:
        :class:`SigmaMinus`

    Parameters:
        a (complex, optional): Scaling factor for the matrix. Defaults to 1.

    Examples:
        >>> from mimiqcircuits import *
        >>> SigmaPlus()
        SigmaPlus(1)

        >>> SigmaPlus(0.5)
        SigmaPlus(0.5)

        >>> c = Circuit()
        >>> c.push(ExpectationValue(SigmaPlus()), 1, 1)
        2-qubit circuit with 1 instructions:
        └── ⟨SigmaPlus(1)⟩ @ q[1], z[1]
        <BLANKLINE>
    """

    _name = "SigmaPlus"
    _num_qubits = 1
    _parnames = ()
    _qregsizes = [1]

    def __init__(self, a=1):
        self.a = a
        super().__init__()
        self._parnames = ("a",)

    @property
    def parnames(self):
        return self._parnames

    @property
    def num_qubits(self):
        return self._num_qubits

    def _matrix(self):
        return se.Matrix([[0, 0], [self.a, 0]])

    def opsquared(self):
        return mc.Projector0(abs(self.a) ** 2)

    def rescale(self, scale):
        return SigmaPlus(self.a * scale)

    def __str__(self):
        return f"{self.opname()}({self.a})"


__all__ = [
    "SigmaMinus",
    "SigmaPlus",
]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc
from mimiqcircuits.matrices import pmatrix
from mimiqcircuits.operations.utils import control_one_defined


class GateP(mc.Gate):
    r"""Single qubit Phase gate.

    **Matrix representation:**

    .. math::
        \operatorname{P}(\lambda) =
        \operatorname{U}(0,0,\lambda) =
        \begin{pmatrix}
            1 & 0 \\
            0 & \mathrm{e}^{i\lambda}
        \end{pmatrix}

    Parameters:
        lambda: Phase angle

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> lmbda = Symbol('lambda')
        >>> GateP(lmbda)
        P(lambda)
        >>> GateP(lmbda).matrix()
        [1.0, 0]
        [0, exp(I*lambda)]
        <BLANKLINE>
        >>> c = Circuit().push(GateP(lmbda), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── P(lambda) @ q[0]
        <BLANKLINE>
        >>> GateP(lmbda).power(2), GateP(lmbda).inverse()
        (P(2*lambda), P(-lambda))
        >>> GateP(lmbda).decompose()
        1-qubit circuit with 1 instructions:
        └── U(0, 0, lambda, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "P"

    _num_qubits = 1
    _qregsizes = [1]

    _parnames = ("lmbda",)

    def __init__(self, lmbda):
        self.lmbda = lmbda

    def _matrix(self):
        return pmatrix(self.lmbda)

    def inverse(self):
        return GateP(-self.lmbda)

    def _power(self, p):
        return GateP(self.lmbda * p)

    def _control(self, n):
        return control_one_defined(
            n, self, mc.GateCP(self.lmbda), mc.GateCCP(self.lmbda)
        )

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits
        circ.push(
            mc.GateU(
                0,
                0,
                self.lmbda,
            ),
            q,
        )
        return circ
